import pytest

from sqlframe.base.catalog import CatalogMetadata, Column, Database, Function, Table
from sqlframe.postgres.session import PostgresSession

pytest_plugins = ["tests.integration.fixtures"]


def test_current_catalog(postgres_session: PostgresSession):
    assert postgres_session.catalog.currentCatalog() == "tests"


def test_set_current_catalog(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.setCurrentCatalog("catalog1")


def test_list_catalogs(postgres_session: PostgresSession):
    assert sorted(postgres_session.catalog.listCatalogs(), key=lambda x: x.name) == [
        CatalogMetadata(name="tests", description=None)
    ]


def test_current_database(postgres_session: PostgresSession):
    assert postgres_session.catalog.currentDatabase() == "public"


def test_set_current_database(postgres_session: PostgresSession):
    assert postgres_session.catalog.currentDatabase() == "public"
    postgres_session.catalog.setCurrentDatabase("db1")
    assert postgres_session.catalog.currentDatabase() == "db1"


def test_list_databases_pattern(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listDatabases("db*"), key=lambda x: (x.catalog, x.name)
    ) == [
        Database(name="db1", catalog="tests", description=None, locationUri=""),
    ]


def test_get_database_no_match(postgres_session: PostgresSession):
    with pytest.raises(ValueError):
        assert postgres_session.catalog.getDatabase("nonexistent")


def test_get_database_name_only(postgres_session: PostgresSession):
    assert postgres_session.catalog.getDatabase("db1") == Database(
        name="db1", catalog="tests", description=None, locationUri=""
    )


def test_get_database_name_and_catalog(postgres_session: PostgresSession):
    # catalog1 is ignored
    assert postgres_session.catalog.getDatabase("catalog1.information_schema") == Database(
        name="information_schema", catalog="tests", description=None, locationUri=""
    )


def test_database_exists_does_exist(postgres_session: PostgresSession):
    assert postgres_session.catalog.databaseExists("db1") is True


def test_database_exists_does_not_exist(postgres_session: PostgresSession):
    assert postgres_session.catalog.databaseExists("nonexistent") is False


def test_list_tables_no_args(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listTables(), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="tests",
            namespace=["public"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_no_catalog(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listTables("public"), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="tests",
            namespace=["public"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_and_catalog(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listTables("tests.public"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="tests",
            namespace=["public"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_pattern(postgres_session: PostgresSession):
    assert Table(
        name="table1",
        catalog="tests",
        namespace=["public"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    ) in postgres_session.catalog.listTables(pattern="tab*")


def test_get_table(postgres_session: PostgresSession):
    assert postgres_session.catalog.getTable("tests.public.table1") == Table(
        name="table1",
        catalog="tests",
        namespace=["public"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    )


def test_get_table_not_exists(postgres_session: PostgresSession):
    with pytest.raises(ValueError):
        assert postgres_session.catalog.getTable("tests.public.nonexistent")


def test_list_functions(postgres_session: PostgresSession):
    assert (
        Function(
            name="testing",
            catalog="tests",
            namespace=["public"],
            description=None,
            className="",
            isTemporary=False,
        )
        in postgres_session.catalog.listFunctions()
    )


def test_list_functions_pattern(postgres_session: PostgresSession):
    assert postgres_session.catalog.listFunctions(pattern="test*") == [
        Function(
            name="testing",
            catalog="tests",
            namespace=["public"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_function_exists_does_exist(postgres_session: PostgresSession):
    assert postgres_session.catalog.functionExists("testing") is True


def test_function_exists_does_not_exist(postgres_session: PostgresSession):
    assert postgres_session.catalog.functionExists("nonexistent") is False


def test_get_function_exists(postgres_session: PostgresSession):
    assert postgres_session.catalog.getFunction("tests.public.testing") == Function(
        name="testing",
        catalog="tests",
        namespace=["public"],
        description=None,
        className="",
        isTemporary=False,
    )


def test_get_function_not_exists(postgres_session: PostgresSession):
    with pytest.raises(ValueError):
        assert postgres_session.catalog.getFunction("default.main.nonexistent")


def test_list_columns(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listColumns("tests.public.table1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_list_columns_use_db_name(postgres_session: PostgresSession):
    assert sorted(
        postgres_session.catalog.listColumns("table1", dbName="tests.public"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_table_exists_table_name_only(postgres_session: PostgresSession):
    assert postgres_session.catalog.tableExists("tests.public.table1") is True


def test_table_exists_table_name_and_db_name(postgres_session: PostgresSession):
    assert postgres_session.catalog.tableExists("table1", dbName="tests.public") is True


def test_table_not_exists(postgres_session: PostgresSession):
    assert postgres_session.catalog.tableExists("nonexistent") is False


def test_create_external_table(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.createExternalTable("table1", "tests.public", "path/to/table")


def test_create_table(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.createTable("table1", "tests.public")


def test_drop_temp_view(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.dropTempView("view1")


def test_drop_global_temp_view(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.dropGlobalTempView("view1")


def test_register_function(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.registerFunction("function1", lambda x: x)


def test_is_cached(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.isCached("table1")


def test_cache_table(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.cacheTable("table1")


def test_uncache_table(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.uncacheTable("table1")


def test_clear_cache(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.clearCache()


def test_refresh_table(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.refreshTable("table1")


def test_recover_partitions(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.recoverPartitions("table1")


def test_refresh_by_path(postgres_session: PostgresSession):
    with pytest.raises(NotImplementedError):
        postgres_session.catalog.refreshByPath("path/to/table")



import pytest

from sqlframe.postgres import Column as PostgresColumn
from sqlframe.postgres import (
    PostgresCatalog,
    PostgresDataFrame,
    PostgresDataFrameNaFunctions,
    PostgresDataFrameReader,
    PostgresDataFrameStatFunctions,
    PostgresDataFrameWriter,
    PostgresGroupedData,
    PostgresSession,
    PostgresUDFRegistration,
)
from sqlframe.postgres import Row as PostgresRow
from sqlframe.postgres import Window as PostgresWindow
from sqlframe.postgres import WindowSpec as PostgresWindowSpec
from sqlframe.postgres import functions as PostgresF
from sqlframe.postgres import types as PostgresTypes


@pytest.mark.forked
def test_activate_postgres(check_pyspark_imports):
    check_pyspark_imports(
        "postgres",
        sqlf_session=PostgresSession,
        sqlf_catalog=PostgresCatalog,
        sqlf_column=PostgresColumn,
        sqlf_dataframe=PostgresDataFrame,
        sqlf_grouped_data=PostgresGroupedData,
        sqlf_window=PostgresWindow,
        sqlf_window_spec=PostgresWindowSpec,
        sqlf_functions=PostgresF,
        sqlf_types=PostgresTypes,
        sqlf_udf_registration=PostgresUDFRegistration,
        sqlf_dataframe_reader=PostgresDataFrameReader,
        sqlf_dataframe_writer=PostgresDataFrameWriter,
        sqlf_dataframe_na_functions=PostgresDataFrameNaFunctions,
        sqlf_dataframe_stat_functions=PostgresDataFrameStatFunctions,
        sqlf_row=PostgresRow,
    )



import inspect
import sys

import sqlframe.base.functions  # noqa

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines") and "databricks" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import (  # noqa
    percentile_without_disc as percentile,
    add_months_by_multiplication as add_months,
    arrays_overlap_renamed as arrays_overlap,
    _is_string_using_typeof_string_lcase as _is_string,
    try_element_at_zero_based as try_element_at,
    get_json_object_using_function as get_json_object,
)



from sqlframe.redshift.catalog import RedshiftCatalog
from sqlframe.redshift.column import Column
from sqlframe.redshift.dataframe import (
    RedshiftDataFrame,
    RedshiftDataFrameNaFunctions,
    RedshiftDataFrameStatFunctions,
)
from sqlframe.redshift.group import RedshiftGroupedData
from sqlframe.redshift.readwriter import (
    RedshiftDataFrameReader,
    RedshiftDataFrameWriter,
)
from sqlframe.redshift.session import RedshiftSession
from sqlframe.redshift.types import Row
from sqlframe.redshift.udf import RedshiftUDFRegistration
from sqlframe.redshift.window import Window, WindowSpec

__all__ = [
    "Column",
    "RedshiftCatalog",
    "RedshiftDataFrame",
    "RedshiftDataFrameNaFunctions",
    "RedshiftGroupedData",
    "RedshiftDataFrameReader",
    "RedshiftDataFrameWriter",
    "RedshiftSession",
    "RedshiftDataFrameStatFunctions",
    "RedshiftUDFRegistration",
    "Row",
    "Window",
    "WindowSpec",
]



import typing as t

import pytest
from sqlglot.errors import OptimizeError

from sqlframe.base.exceptions import TableSchemaError
from sqlframe.standalone import functions as F
from sqlframe.standalone.session import StandaloneSession

pytest_plugins = ["tests.common_fixtures", "tests.unit.standalone.fixtures"]


tests = [
    (
        "All lower no intention of CS",
        "test",
        "test",
        {"name": "VARCHAR"},
        "name",
        '''SELECT "TEST"."NAME" AS "NAME" FROM "TEST" AS "TEST"''',
    ),
    (
        "Table has CS while column does not",
        '"Test"',
        '"Test"',
        {"name": "VARCHAR"},
        "name",
        '''SELECT "Test"."NAME" AS "NAME" FROM "Test" AS "Test"''',
    ),
    (
        "Column has CS while table does not",
        "test",
        "test",
        {'"Name"': "VARCHAR"},
        '"Name"',
        '''SELECT "TEST"."Name" AS "Name" FROM "TEST" AS "TEST"''',
    ),
    (
        "Both Table and column have CS",
        '"Test"',
        '"Test"',
        {'"Name"': "VARCHAR"},
        '"Name"',
        '''SELECT "Test"."Name" AS "Name" FROM "Test" AS "Test"''',
    ),
    (
        "Lowercase CS table and column",
        '"test"',
        '"test"',
        {'"name"': "VARCHAR"},
        '"name"',
        '''SELECT "test"."name" AS "name" FROM "test" AS "test"''',
    ),
    (
        "CS table and column and query table but no CS in query column",
        '"test"',
        '"test"',
        {'"name"': "VARCHAR"},
        "name",
        OptimizeError(),
    ),
    (
        "CS table and column and query column but no CS in query table",
        '"test"',
        "test",
        {'"name"': "VARCHAR"},
        '"name"',
        OptimizeError(),
    ),
]


@pytest.mark.parametrize(
    "table_name, spark_table, schema, spark_column, expected",
    [test[1:] for test in tests],
    ids=[test[0] for test in tests],
)
def test_basic_case_sensitivity(
    sqlf_sf: StandaloneSession,
    compare_sql: t.Callable,
    table_name,
    spark_table,
    schema,
    spark_column,
    expected,
):
    sqlf_sf.catalog.add_table(table_name, schema)
    if isinstance(expected, OptimizeError):
        with pytest.raises((OptimizeError, TableSchemaError)):
            df = sqlf_sf.table(spark_table).select(F.col(spark_column))
            df.sql()
    else:
        df = sqlf_sf.table(spark_table).select(F.col(spark_column))
        compare_sql(df, expected)


@pytest.mark.parametrize(
    "col_func, expected",
    [
        (lambda x: x, '"Name"'),
        (lambda x: x.alias("nAME"), '"Name" AS NAME'),
        (lambda x: x.alias('"nAME"'), '"Name" AS "nAME"'),
    ],
)
def test_alias(
    sqlf_sf: StandaloneSession, compare_sql: t.Callable, col_func: t.Callable, expected: str
):
    # Need to run `F.col` after the Snowflake SQLFrame session has been set since it will be default dialect otherwise
    col = col_func(F.col('"Name"'))
    compare_sql(col, expected, dialect=sqlf_sf.input_dialect)



import typing as t

import pytest

from sqlframe.standalone import functions as F
from sqlframe.standalone import types
from sqlframe.standalone.session import StandaloneSession

pytest_plugins = ["tests.common_fixtures", "tests.unit.standalone.fixtures"]


def test_cdf_one_row(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[1, 2]], ["cola", "colb"])
    expected = "SELECT CAST(`a1`.`cola` AS BIGINT) AS `cola`, CAST(`a1`.`colb` AS BIGINT) AS `colb` FROM VALUES (1, 2) AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_cdf_multiple_rows(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[1, 2], [3, 4], [None, 6]], ["cola", "colb"])
    expected = "SELECT CAST(`a1`.`cola` AS BIGINT) AS `cola`, CAST(`a1`.`colb` AS BIGINT) AS `colb` FROM VALUES (1, 2), (3, 4), (NULL, 6) AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_cdf_no_schema(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[1, 2], [3, 4], [None, 6]])
    expected = "SELECT CAST(`a1`.`_1` AS BIGINT) AS `_1`, CAST(`a1`.`_2` AS BIGINT) AS `_2` FROM VALUES (1, 2), (3, 4), (NULL, 6) AS `a1`(`_1`, `_2`)"
    compare_sql(df, expected)


def test_cdf_row_mixed_primitives(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[1, 10.1, "test", False, None]])
    expected = "SELECT CAST(`a1`.`_1` AS BIGINT) AS `_1`, `a1`.`_2` AS `_2`, CAST(`a1`.`_3` AS STRING) AS `_3`, `a1`.`_4` AS `_4`, `a1`.`_5` AS `_5` FROM VALUES (1, 10.1, 'test', FALSE, NULL) AS `a1`(`_1`, `_2`, `_3`, `_4`, `_5`)"
    compare_sql(df, expected)


def test_cdf_dict_rows(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame(
        [{"cola": 1, "colb": "test"}, {"cola": 2, "colb": "test2"}]
    )
    expected = "SELECT CAST(`a1`.`cola` AS BIGINT) AS `cola`, CAST(`a1`.`colb` AS STRING) AS `colb` FROM VALUES (1, 'test'), (2, 'test2') AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_cdf_str_schema(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[1, "test"]], "cola INT, colb STRING")
    expected = "SELECT `a1`.`cola` AS `cola`, CAST(`a1`.`colb` AS STRING) AS `colb` FROM VALUES (1, 'test') AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_typed_schema_basic(standalone_session: StandaloneSession, compare_sql: t.Callable):
    schema = types.StructType(
        [
            types.StructField("cola", types.IntegerType()),
            types.StructField("colb", types.StringType()),
        ]
    )
    df = standalone_session.createDataFrame([[1, "test"]], schema)
    expected = "SELECT `a1`.`cola` AS `cola`, CAST(`a1`.`colb` AS STRING) AS `colb` FROM VALUES (1, 'test') AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_single_value_rows(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([1, 2, 3], "int")
    expected = "SELECT `a1`.`value` AS `value` FROM VALUES (1), (2), (3) AS `a1`(`value`)"
    compare_sql(df, expected)


def test_row_value(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([types.Row(cola=1, colb="test")])  # type: ignore
    expected = "SELECT CAST(`a1`.`cola` AS BIGINT) AS `cola`, CAST(`a1`.`colb` AS STRING) AS `colb` FROM VALUES (1, 'test') AS `a1`(`cola`, `colb`)"
    compare_sql(df, expected)


def test_typed_schema_nested_map(standalone_session: StandaloneSession, compare_sql: t.Callable):
    schema = types.StructType(
        [
            types.StructField(
                "cola",
                types.MapType(
                    keyType=types.IntegerType(),
                    valueType=types.StringType(),
                ),
            )
        ]
    )
    df = standalone_session.createDataFrame([[{"sub_cola": 1, "sub_colb": "test"}]], schema)  # type: ignore
    expected = "SELECT `a1`.`cola` AS `cola` FROM VALUES (MAP('sub_cola', 1, 'sub_colb', 'test')) AS `a1`(`cola`)"

    compare_sql(df, expected)


def test_nested_struct(standalone_session: StandaloneSession, compare_sql: t.Callable):
    df = standalone_session.createDataFrame([[types.Row(cola=1, colb="test")]])  # type: ignore
    expected = (
        "SELECT `a1`.`_1` AS `_1` FROM VALUES (STRUCT(1 AS `cola`, 'test' AS `colb`)) AS `a1`(`_1`)"
    )
    compare_sql(df, expected)


def test_sql_select_only(standalone_session: StandaloneSession, compare_sql: t.Callable):
    query = "SELECT cola, colb FROM table"
    standalone_session.catalog.add_table("table", {"cola": "string", "colb": "string"})
    df = standalone_session.sql(query)
    compare_sql(
        df, "SELECT `table`.`cola` AS `cola`, `table`.`colb` AS `colb` FROM `table` AS `table`"
    )


def test_sql_with_aggs(standalone_session: StandaloneSession, compare_sql: t.Callable):
    query = "SELECT cola, colb FROM table"
    standalone_session.catalog.add_table("table", {"cola": "string", "colb": "string"})
    df = standalone_session.sql(query).groupBy(F.col("cola")).agg(F.sum("colb"))
    compare_sql(
        df,
        "WITH `t34970089` AS (SELECT `table`.`cola` AS `cola`, `table`.`colb` AS `colb` FROM `table` AS `table`), `t50917871` AS (SELECT `cola`, `colb` FROM `t34970089`) SELECT `cola`, SUM(`colb`) FROM `t50917871` GROUP BY `cola`",
        pretty=False,
        optimize=False,
    )


def test_sql_create(standalone_session: StandaloneSession, compare_sql: t.Callable):
    query = "CREATE TABLE new_table AS WITH t1 AS (SELECT cola, colb FROM table) SELECT cola, colb, FROM t1"
    standalone_session.catalog.add_table("table", {"cola": "string", "colb": "string"})
    df = standalone_session.sql(query)
    expected = "CREATE TABLE `new_table` AS SELECT `table`.`cola` AS `cola`, `table`.`colb` AS `colb` FROM `table` AS `table`"
    compare_sql(df, expected)


def test_sql_insert(standalone_session: StandaloneSession, compare_sql: t.Callable):
    query = (
        "WITH t1 AS (SELECT cola, colb FROM table) INSERT INTO new_table SELECT cola, colb FROM t1"
    )
    standalone_session.catalog.add_table("table", {"cola": "string", "colb": "string"})
    df = standalone_session.sql(query)
    expected = "INSERT INTO `new_table` SELECT `table`.`cola` AS `cola`, `table`.`colb` AS `colb` FROM `table` AS `table`"
    compare_sql(df, expected)


def test_session_create_builder_patterns():
    assert StandaloneSession.builder.appName("abc").getOrCreate() == StandaloneSession()


# @pytest.mark.parametrize(
#     "input, expected",
#     [
#         (
#             StandaloneSession._to_row(["a"], [1]),
#             types.Row(a=1),
#         ),
#         (
#             StandaloneSession._to_row(["a", "b"], [1, 2]),
#             types.Row(a=1, b=2),
#         ),
#         (
#             StandaloneSession._to_row(["a", "a"], [1, 2]),
#             types.Row(a=1, a=2),
#         ),
#     ],
# )
# def test_to_row(input, expected):
#     assert input == expected



import typing as t

from sqlframe.base.column import Column as Column
from sqlframe.base.function_alternatives import (  # noqa
    any_value_ignore_nulls_not_supported as any_value,
)
from sqlframe.base.function_alternatives import (
    array_union_using_array_concat as array_union,
)
from sqlframe.base.function_alternatives import (
    base64_from_blob as base64,
)
from sqlframe.base.function_alternatives import (
    bit_length_from_length as bit_length,
)
from sqlframe.base.function_alternatives import (
    collect_set_from_list_distinct as collect_set,
)
from sqlframe.base.function_alternatives import (
    concat_ws_from_array_to_string as concat_ws,
)
from sqlframe.base.function_alternatives import (
    current_user_from_session_user as current_user,
)
from sqlframe.base.function_alternatives import (
    dayofmonth_from_extract_with_day as dayofmonth,
)
from sqlframe.base.function_alternatives import (
    dayofweek_from_extract as dayofweek,
)
from sqlframe.base.function_alternatives import (
    dayofyear_from_extract as dayofyear,
)
from sqlframe.base.function_alternatives import (
    e_literal as e,
)
from sqlframe.base.function_alternatives import (
    element_at_using_brackets as element_at,
)
from sqlframe.base.function_alternatives import (
    expm1_from_exp as expm1,
)
from sqlframe.base.function_alternatives import (
    extract_convert_to_var as extract,
)
from sqlframe.base.function_alternatives import (
    factorial_from_case_statement as factorial,
)
from sqlframe.base.function_alternatives import (
    format_string_with_format as format_string,
)
from sqlframe.base.function_alternatives import (
    hash_from_farm_fingerprint as hash,
)
from sqlframe.base.function_alternatives import (
    hex_casted_as_bytes as hex,
)
from sqlframe.base.function_alternatives import (
    hour_from_extract as hour,
)
from sqlframe.base.function_alternatives import (
    instr_using_strpos as instr,
)
from sqlframe.base.function_alternatives import (
    isnull_using_equal as isnull,
)
from sqlframe.base.function_alternatives import (
    last_day_with_cast as last_day,
)
from sqlframe.base.function_alternatives import (
    log1p_from_log as log1p,
)
from sqlframe.base.function_alternatives import (
    make_date_from_date_func as make_date,
)
from sqlframe.base.function_alternatives import (
    minute_from_extract as minute,
)
from sqlframe.base.function_alternatives import (
    month_from_extract as month,
)
from sqlframe.base.function_alternatives import (
    nanvl_as_case as nanvl,
)
from sqlframe.base.function_alternatives import (
    overlay_from_substr as overlay,
)
from sqlframe.base.function_alternatives import (
    percentile_approx_without_accuracy_and_plural as percentile_approx,
)
from sqlframe.base.function_alternatives import (
    position_as_strpos as position,
)
from sqlframe.base.function_alternatives import (
    quarter_from_extract as quarter,
)
from sqlframe.base.function_alternatives import (
    rand_no_seed as rand,
)
from sqlframe.base.function_alternatives import (
    regexp_extract_only_one_group as regexp_extract,
)
from sqlframe.base.function_alternatives import (
    rint_from_round as rint,
)
from sqlframe.base.function_alternatives import (
    second_from_extract as second,
)
from sqlframe.base.function_alternatives import (
    sequence_from_generate_array as sequence,
)
from sqlframe.base.function_alternatives import (
    sha1_force_sha1_and_to_hex as sha1,
)
from sqlframe.base.function_alternatives import (
    split_with_split as split,
)
from sqlframe.base.function_alternatives import (
    to_date_from_timestamp as to_date,
)
from sqlframe.base.function_alternatives import (
    weekofyear_from_extract_as_isoweek as weekofyear,
)
from sqlframe.base.function_alternatives import (
    year_from_extract as year,
)
from sqlframe.base.functions import abs as abs
from sqlframe.base.functions import acos as acos
from sqlframe.base.functions import acosh as acosh
from sqlframe.base.functions import add_months as add_months
from sqlframe.base.functions import approx_count_distinct as approx_count_distinct
from sqlframe.base.functions import approxCountDistinct as approxCountDistinct
from sqlframe.base.functions import array as array
from sqlframe.base.functions import array_contains as array_contains
from sqlframe.base.functions import array_join as array_join
from sqlframe.base.functions import asc as asc
from sqlframe.base.functions import asc_nulls_first as asc_nulls_first
from sqlframe.base.functions import asc_nulls_last as asc_nulls_last
from sqlframe.base.functions import ascii as ascii
from sqlframe.base.functions import asin as asin
from sqlframe.base.functions import asinh as asinh
from sqlframe.base.functions import atan as atan
from sqlframe.base.functions import atan2 as atan2
from sqlframe.base.functions import atanh as atanh
from sqlframe.base.functions import avg as avg
from sqlframe.base.functions import bitwise_not as bitwise_not
from sqlframe.base.functions import bitwiseNOT as bitwiseNOT
from sqlframe.base.functions import bool_and as bool_and
from sqlframe.base.functions import bool_or as bool_or
from sqlframe.base.functions import call_function as call_function
from sqlframe.base.functions import cbrt as cbrt
from sqlframe.base.functions import ceil as ceil
from sqlframe.base.functions import ceiling as ceiling
from sqlframe.base.functions import char as char
from sqlframe.base.functions import coalesce as coalesce
from sqlframe.base.functions import col as col
from sqlframe.base.functions import collect_list as collect_list
from sqlframe.base.functions import concat as concat
from sqlframe.base.functions import corr as corr
from sqlframe.base.functions import cos as cos
from sqlframe.base.functions import cosh as cosh
from sqlframe.base.functions import cot as cot
from sqlframe.base.functions import count as count
from sqlframe.base.functions import count_if as count_if
from sqlframe.base.functions import covar_pop as covar_pop
from sqlframe.base.functions import covar_samp as covar_samp
from sqlframe.base.functions import csc as csc
from sqlframe.base.functions import cume_dist as cume_dist
from sqlframe.base.functions import current_date as current_date
from sqlframe.base.functions import current_timestamp as current_timestamp
from sqlframe.base.functions import date_add as date_add
from sqlframe.base.functions import date_diff as date_diff
from sqlframe.base.functions import date_format as date_format
from sqlframe.base.functions import date_sub as date_sub
from sqlframe.base.functions import date_trunc as date_trunc
from sqlframe.base.functions import dateadd as dateadd
from sqlframe.base.functions import datediff as datediff
from sqlframe.base.functions import dense_rank as dense_rank
from sqlframe.base.functions import desc as desc
from sqlframe.base.functions import desc_nulls_first as desc_nulls_first
from sqlframe.base.functions import desc_nulls_last as desc_nulls_last
from sqlframe.base.functions import exp as exp
from sqlframe.base.functions import explode as explode
from sqlframe.base.functions import explode_outer as explode_outer
from sqlframe.base.functions import expr as expr
from sqlframe.base.functions import floor as floor
from sqlframe.base.functions import get_json_object as get_json_object
from sqlframe.base.functions import greatest as greatest
from sqlframe.base.functions import ifnull as ifnull
from sqlframe.base.functions import initcap as initcap
from sqlframe.base.functions import input_file_name as input_file_name
from sqlframe.base.functions import isnan as isnan
from sqlframe.base.functions import lag as lag
from sqlframe.base.functions import lcase as lcase
from sqlframe.base.functions import lead as lead
from sqlframe.base.functions import least as least
from sqlframe.base.functions import left as left
from sqlframe.base.functions import length as length
from sqlframe.base.functions import lit as lit
from sqlframe.base.functions import ln as ln
from sqlframe.base.functions import log as log
from sqlframe.base.functions import log2 as log2
from sqlframe.base.functions import log10 as log10
from sqlframe.base.functions import lower as lower
from sqlframe.base.functions import lpad as lpad
from sqlframe.base.functions import ltrim as ltrim
from sqlframe.base.functions import max as max
from sqlframe.base.functions import max_by as max_by
from sqlframe.base.functions import md5 as md5
from sqlframe.base.functions import mean as mean
from sqlframe.base.functions import min as min
from sqlframe.base.functions import min_by as min_by
from sqlframe.base.functions import now as now
from sqlframe.base.functions import nth_value as nth_value
from sqlframe.base.functions import ntile as ntile
from sqlframe.base.functions import nullif as nullif
from sqlframe.base.functions import nvl as nvl
from sqlframe.base.functions import nvl2 as nvl2
from sqlframe.base.functions import octet_length as octet_length
from sqlframe.base.functions import percent_rank as percent_rank
from sqlframe.base.functions import posexplode as posexplode
from sqlframe.base.functions import posexplode_outer as posexplode_outer
from sqlframe.base.functions import pow as pow
from sqlframe.base.functions import power as power
from sqlframe.base.functions import rank as rank
from sqlframe.base.functions import regexp_like as regexp_like
from sqlframe.base.functions import regexp_replace as regexp_replace
from sqlframe.base.functions import repeat as repeat
from sqlframe.base.functions import reverse as reverse
from sqlframe.base.functions import right as right
from sqlframe.base.functions import rlike as rlike
from sqlframe.base.functions import round as round
from sqlframe.base.functions import row_number as row_number
from sqlframe.base.functions import rpad as rpad
from sqlframe.base.functions import rtrim as rtrim
from sqlframe.base.functions import sec as sec
from sqlframe.base.functions import sha as sha
from sqlframe.base.functions import shiftLeft as shiftLeft
from sqlframe.base.functions import shiftleft as shiftleft
from sqlframe.base.functions import shiftRight as shiftRight
from sqlframe.base.functions import shiftright as shiftright
from sqlframe.base.functions import sign as sign
from sqlframe.base.functions import signum as signum
from sqlframe.base.functions import sin as sin
from sqlframe.base.functions import sinh as sinh
from sqlframe.base.functions import size as size
from sqlframe.base.functions import soundex as soundex
from sqlframe.base.functions import sqrt as sqrt
from sqlframe.base.functions import startswith as startswith
from sqlframe.base.functions import stddev as stddev
from sqlframe.base.functions import stddev_pop as stddev_pop
from sqlframe.base.functions import stddev_samp as stddev_samp
from sqlframe.base.functions import struct as struct
from sqlframe.base.functions import substr as substr
from sqlframe.base.functions import substring as substring
from sqlframe.base.functions import sum as sum
from sqlframe.base.functions import sum_distinct as sum_distinct
from sqlframe.base.functions import sumDistinct as sumDistinct
from sqlframe.base.functions import tan as tan
from sqlframe.base.functions import tanh as tanh
from sqlframe.base.functions import timestamp_seconds as timestamp_seconds
from sqlframe.base.functions import to_timestamp as to_timestamp
from sqlframe.base.functions import toDegrees as toDegrees
from sqlframe.base.functions import toRadians as toRadians
from sqlframe.base.functions import translate as translate
from sqlframe.base.functions import trim as trim
from sqlframe.base.functions import trunc as trunc
from sqlframe.base.functions import ucase as ucase
from sqlframe.base.functions import unbase64 as unbase64
from sqlframe.base.functions import unhex as unhex
from sqlframe.base.functions import upper as upper
from sqlframe.base.functions import user as user
from sqlframe.base.functions import var_pop as var_pop
from sqlframe.base.functions import var_samp as var_samp
from sqlframe.base.functions import variance as variance
from sqlframe.base.functions import when as when
from sqlframe.base.util import get_func_from_session as get_func_from_session

if t.TYPE_CHECKING:
    from sqlframe.base._typing import ColumnOrLiteral, ColumnOrName

def array_distinct(col: ColumnOrName) -> Column: ...
def array_max(col: ColumnOrName) -> Column: ...
def array_min(col: ColumnOrName) -> Column: ...
def array_position(col: ColumnOrName, value: ColumnOrLiteral) -> Column: ...
def array_remove(col: ColumnOrName, value: ColumnOrLiteral) -> Column: ...
def array_sort(col: ColumnOrName, asc: t.Optional[bool] = ...) -> Column: ...
def bin(col: ColumnOrName) -> Column: ...
def bround(col: ColumnOrName, scale: t.Optional[int] = ...) -> Column: ...
def degrees(col: ColumnOrName) -> Column: ...
def format_number(col: ColumnOrName, d: int) -> Column: ...
def from_unixtime(col: ColumnOrName, format: t.Optional[str] = ...) -> Column: ...
def months_between(
    date1: ColumnOrName, date2: ColumnOrName, roundOff: t.Optional[bool] = ...
) -> Column: ...
def next_day(col: ColumnOrName, dayOfWeek: str) -> Column: ...
def radians(col: ColumnOrName) -> Column: ...
def slice(
    x: ColumnOrName, start: t.Union[ColumnOrName, int], length: t.Union[ColumnOrName, int]
) -> Column: ...
def sort_array(col: ColumnOrName, asc: t.Optional[bool] = ...) -> Column: ...
def substring_index(str: ColumnOrName, delim: str, count: int) -> Column: ...
def typeof(col: ColumnOrName) -> Column: ...
def unix_timestamp(
    timestamp: t.Optional[ColumnOrName] = ..., format: t.Optional[str] = ...
) -> Column: ...



import typing as t

from sqlframe.base.catalog import _BaseCatalog

if t.TYPE_CHECKING:
    from sqlframe.standalone.dataframe import StandaloneDataFrame
    from sqlframe.standalone.session import StandaloneSession


class StandaloneCatalog(_BaseCatalog["StandaloneSession", "StandaloneDataFrame"]):
    """User-facing catalog API, accessible through `SparkSession.catalog`."""

    pass



from __future__ import annotations

import importlib
import typing as t
import unicodedata

from sqlglot import expressions as exp
from sqlglot import parse_one, to_table
from sqlglot.dialects import DuckDB
from sqlglot.dialects.dialect import Dialect, DialectType
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
from sqlglot.optimizer.qualify_columns import (
    quote_identifiers as quote_identifiers_func,
)
from sqlglot.schema import ensure_column_mapping as sqlglot_ensure_column_mapping

if t.TYPE_CHECKING:
    from pandas.core.frame import DataFrame as PandasDataFrame
    from pyspark.sql.dataframe import SparkSession as PySparkSession

    from sqlframe.base import types
    from sqlframe.base._typing import (
        ColumnOrLiteral,
        OptionalPrimitiveType,
        SchemaInput,
    )
    from sqlframe.base.column import Column
    from sqlframe.base.session import _BaseSession
    from sqlframe.base.types import StructType


def decoded_str(value: t.Union[str, bytes]) -> str:
    if isinstance(value, bytes):
        return value.decode("utf-8")
    return value


def schema_(
    db: exp.Identifier | str,
    catalog: t.Optional[exp.Identifier | str] = None,
    quoted: t.Optional[bool] = None,
) -> exp.Table:
    """Build a Schema.

    Args:
        db: Database name.
        catalog: Catalog name.
        quoted: Whether to force quotes on the schema's identifiers.

    Returns:
        The new Schema instance.
    """
    return exp.Table(
        this=None,
        db=exp.to_identifier(db, quoted=quoted) if db else None,
        catalog=exp.to_identifier(catalog, quoted=quoted) if catalog else None,
    )


def to_schema(
    sql_path: t.Union[str, exp.Table], dialect: t.Optional[DialectType] = None
) -> exp.Table:
    if isinstance(sql_path, exp.Table) and sql_path.this is None:
        return sql_path
    table = exp.to_table(
        sql_path.copy() if isinstance(sql_path, exp.Table) else sql_path, dialect=dialect
    )
    table.set("catalog", table.args.get("db"))
    table.set("db", table.args.get("this"))
    table.set("this", None)
    return table


def get_column_mapping_from_schema_input(
    schema: SchemaInput, dialect: DialectType = None
) -> t.Dict[str, t.Optional[exp.DataType]]:
    from sqlframe.base import types

    if isinstance(schema, dict):
        value = schema
    elif isinstance(schema, str):
        col_name_type_strs = [x.strip() for x in schema.split(",")]
        if len(col_name_type_strs) == 1 and len(col_name_type_strs[0].split(" ")) == 1:
            value = {"value": col_name_type_strs[0].strip()}
        elif schema.startswith("struct<") and schema.endswith(">"):
            value = {
                name_type_str.split(":")[0].strip(): name_type_str.split(":")[1].strip()
                for name_type_str in schema[7:-1].split(",")
            }
        else:
            value = {
                name_type_str.split(" ")[0].strip(): name_type_str.split(" ")[1].strip()
                for name_type_str in col_name_type_strs
            }
    elif isinstance(schema, types.StructType):
        value = {struct_field.name: struct_field.dataType.simpleString() for struct_field in schema}
    else:
        value = {x.strip(): None for x in schema}
    return {
        exp.to_column(k).sql(dialect=dialect): exp.DataType.build(v, dialect=dialect)
        if v is not None
        else v
        for k, v in value.items()
    }
    # return {x.strip(): None for x in schema}  # type: ignore


def get_tables_from_expression_with_join(expression: exp.Select) -> t.List[exp.Table]:
    if not expression.args.get("joins"):
        return []

    left_table = expression.args["from"].this
    other_tables = [join.this for join in expression.args["joins"]]
    return [left_table] + other_tables


def to_csv(options: t.Dict[str, OptionalPrimitiveType], equality_char: str = "=") -> str:
    return ", ".join(
        [f"{k}{equality_char}{v}" for k, v in (options or {}).items() if v is not None]
    )


def ensure_column_mapping(schema: t.Union[str, StructType]) -> t.Dict:
    if isinstance(schema, str):
        col_name_type_strs = [x.strip() for x in schema.split(",")]
        schema = {  # type: ignore
            name_type_str.split(" ")[0].strip(): name_type_str.split(" ")[1].strip()
            for name_type_str in col_name_type_strs
        }
    # TODO: Make a protocol with a `simpleString` attribute as what it looks for instead of the actual
    # `StructType` object.
    elif hasattr(schema, "simpleString"):
        return {struct_field.name: struct_field.dataType.simpleString() for struct_field in schema}
    return sqlglot_ensure_column_mapping(schema)  # type: ignore


# SO: https://stackoverflow.com/questions/37513355/converting-pandas-dataframe-into-spark-dataframe-error
def get_equivalent_spark_type(pandas_type) -> types.DataType:
    """
    This method will retrieve the corresponding spark type given a pandas
    type.

    Args:
        pandas_type (str): pandas data type

    Returns:
        spark data type
    """
    from sqlframe.base import types

    type_map = {
        "datetime64[ns]": types.TimestampType(),
        "int64": types.LongType(),
        "int32": types.IntegerType(),
        "float64": types.DoubleType(),
        "float32": types.FloatType(),
    }
    return type_map.get(str(pandas_type).lower(), types.StringType())


def pandas_to_spark_schema(pandas_df: PandasDataFrame) -> types.StructType:
    """
    This method will return a spark dataframe schema given a pandas dataframe.

    Args:
        pandas_df (pandas.core.frame.DataFrame): pandas DataFrame

    Returns:
        equivalent spark DataFrame schema
    """
    from sqlframe.base import types

    columns = list(
        [
            x.replace("?column?", f"unknown_column_{i}").replace("NULL", f"unknown_column_{i}")
            for i, x in enumerate(pandas_df.columns)
        ]
    )
    d_types = list(pandas_df.dtypes)
    p_schema = types.StructType(
        [
            types.StructField(column, get_equivalent_spark_type(pandas_type))
            for column, pandas_type in zip(columns, d_types)
        ]
    )
    return p_schema


def dialect_to_string(dialect: Dialect) -> str:
    mapping = {v: k for k, v in Dialect.classes.items()}
    return mapping[type(dialect)]


def get_func_from_session(
    name: str,
    session: t.Optional[t.Union[_BaseSession, PySparkSession]] = None,
    fallback: bool = True,
) -> t.Callable:
    from sqlframe.base.session import _BaseSession

    session = session if session else _BaseSession()

    if isinstance(session, _BaseSession):
        dialect_str = dialect_to_string(session.execution_dialect)
        import_path = f"sqlframe.{dialect_str}.functions"
    else:
        import_path = "pyspark.sql.functions"
    try:
        func = getattr(importlib.import_module(import_path), name)
    except AttributeError as e:
        if not fallback:
            raise e
        func = getattr(importlib.import_module("sqlframe.base.functions"), name)
        if session.execution_dialect in func.unsupported_engines:  # type: ignore
            raise NotImplementedError(
                f"{name} is not supported by the engine: {session.execution_dialect}"  # type: ignore
            )
    return func


def soundex(s):
    if not s:
        return ""

    s = unicodedata.normalize("NFKD", s)
    s = s.upper()

    replacements = (
        ("BFPV", "1"),
        ("CGJKQSXZ", "2"),
        ("DT", "3"),
        ("L", "4"),
        ("MN", "5"),
        ("R", "6"),
    )
    result = [s[0]]
    count = 1

    # find would-be replacment for first character
    for lset, sub in replacements:
        if s[0] in lset:
            last = sub
            break
    else:
        last = None

    for letter in s[1:]:
        for lset, sub in replacements:
            if letter in lset:
                if sub != last:
                    result.append(sub)
                    count += 1
                last = sub
                break
        else:
            if letter != "H" and letter != "W":
                # leave last alone if middle letter is H or W
                last = None
        if count == 4:
            break

    result += "0" * (4 - count)
    return "".join(result)


def verify_pandas_installed():
    try:
        import pandas  # noqa
    except ImportError:
        raise ImportError(
            """Pandas is required for this functionality. `pip install "sqlframe[pandas]"` (also include your engine if needed) to install pandas."""
        )


def verify_openai_installed():
    try:
        import openai  # noqa
    except ImportError:
        raise ImportError(
            """OpenAI is required for this functionality. `pip install "sqlframe[openai]"` (also include your engine if needed) to install openai."""
        )


def verify_numpy_installed():
    try:
        import numpy  # noqa
    except ImportError:
        raise ImportError(
            """Numpy is required for this functionality. `pip install "sqlframe[pandas]"` (also include your engine if needed) to install pandas/numpy."""
        )


def quote_preserving_alias_or_name(col: t.Union[exp.Column, exp.Alias]) -> str:
    from sqlframe.base.session import _BaseSession

    if isinstance(col, exp.Alias):
        col = col.args["alias"]
    if isinstance(col, exp.Column):
        col = col.copy()
        col.set("table", None)
    if isinstance(col, (exp.Identifier, exp.Column)):
        return col.sql(dialect=_BaseSession().input_dialect)
    # We may get things like `Null()` expression or maybe literals so we just return the alias or name in those cases
    return col.alias_or_name


def sqlglot_to_spark(sqlglot_dtype: exp.DataType) -> types.DataType:
    from sqlframe.base import types

    primitive_mapping = {
        exp.DataType.Type.VARCHAR: types.VarcharType,
        exp.DataType.Type.CHAR: types.CharType,
        exp.DataType.Type.TEXT: types.StringType,
        exp.DataType.Type.BINARY: types.BinaryType,
        exp.DataType.Type.BOOLEAN: types.BooleanType,
        exp.DataType.Type.INT: types.IntegerType,
        exp.DataType.Type.BIGINT: types.LongType,
        exp.DataType.Type.SMALLINT: types.ShortType,
        exp.DataType.Type.FLOAT: types.FloatType,
        exp.DataType.Type.DOUBLE: types.DoubleType,
        exp.DataType.Type.DECIMAL: types.DecimalType,
        exp.DataType.Type.DATETIME: types.TimestampType,
        exp.DataType.Type.TIMESTAMP: types.TimestampType,
        exp.DataType.Type.TIMESTAMPTZ: types.TimestampType,
        exp.DataType.Type.TIMESTAMPLTZ: types.TimestampType,
        exp.DataType.Type.TIMESTAMPNTZ: types.TimestampType,
        exp.DataType.Type.DATE: types.DateType,
        exp.DataType.Type.JSON: types.StringType,
    }
    if sqlglot_dtype.this in primitive_mapping:
        pyspark_class = primitive_mapping[sqlglot_dtype.this]
        if issubclass(pyspark_class, types.DataTypeWithLength) and sqlglot_dtype.expressions:
            return pyspark_class(length=int(sqlglot_dtype.expressions[0].this.this))
        elif issubclass(pyspark_class, types.DecimalType) and sqlglot_dtype.expressions:
            return pyspark_class(
                precision=int(sqlglot_dtype.expressions[0].this.this),
                scale=int(sqlglot_dtype.expressions[1].this.this),
            )
        return pyspark_class()
    if sqlglot_dtype.this == exp.DataType.Type.ARRAY:
        return types.ArrayType(sqlglot_to_spark(sqlglot_dtype.expressions[0]))
    elif sqlglot_dtype.this == exp.DataType.Type.MAP:
        return types.MapType(
            sqlglot_to_spark(sqlglot_dtype.expressions[0]),
            sqlglot_to_spark(sqlglot_dtype.expressions[1]),
        )
    elif sqlglot_dtype.this in (exp.DataType.Type.STRUCT, exp.DataType.Type.OBJECT):
        return types.StructType(
            [
                types.StructField(
                    name=field.this.alias_or_name,
                    dataType=sqlglot_to_spark(field.args["kind"]),
                )
                for field in sqlglot_dtype.expressions
            ]
        )
    raise NotImplementedError(f"Unsupported data type: {sqlglot_dtype}")


def normalize_string(
    value: t.Union[str, exp.Expression],
    from_dialect: DialectType = None,
    to_dialect: DialectType = None,
    is_pattern: bool = False,
    is_schema: bool = False,
    is_table: bool = False,
    is_datatype: bool = False,
    is_column: bool = False,
    is_query: bool = False,
    to_string_literal: bool = False,
    quote_identifiers: bool = False,
    pretty: bool = False,
) -> str:
    from sqlframe.base.session import _BaseSession

    data_type_replacement_mapping = {
        DuckDB: {
            "TIMESTAMP_NS": "TIMESTAMP",
        }
    }

    session: _BaseSession = _BaseSession()

    str_to_dialect = {
        "input": session.input_dialect,
        "output": session.output_dialect,
        "execution": session.execution_dialect,
    }
    if not to_dialect:
        to_dialect = from_dialect
    from_dialect = str_to_dialect[from_dialect] if isinstance(from_dialect, str) else from_dialect
    to_dialect = str_to_dialect[to_dialect] if isinstance(to_dialect, str) else to_dialect
    if isinstance(value, str):
        assert from_dialect is not None
        if is_pattern:
            star_positions = [i for i, c in enumerate(value) if c == "*"]
            value_without_star = value.replace("*", "")
        else:
            star_positions = []
            value_without_star = value
        value_expression: exp.Expression
        if is_schema:
            value_expression = to_schema(value_without_star, dialect=from_dialect)
        elif is_table:
            value_expression = to_table(value_without_star, dialect=from_dialect)
        elif is_datatype:
            value_without_star = data_type_replacement_mapping.get(from_dialect, {}).get(  # type: ignore
                value_without_star, value_without_star
            )
            value_expression = exp.DataType.build(value_without_star, dialect=from_dialect)
        elif is_column:
            value_expression = exp.to_column(value_without_star, dialect=from_dialect)
        elif is_query:
            value_expression = parse_one(value, dialect=from_dialect)
        else:
            value_expression = exp.parse_identifier(value_without_star, dialect=from_dialect)
    elif isinstance(value, exp.Expression):
        star_positions = []
        value_expression = value.copy()
        value_expression = normalize_identifiers(value_expression, dialect=from_dialect)
    else:
        raise ValueError(f"Unsupported value type: {type(value)}")
    normalized_expression = normalize_identifiers(value_expression, dialect=to_dialect)
    if to_string_literal:
        return normalized_expression.this
    if quote_identifiers:
        quote_identifiers_func(normalized_expression, dialect=to_dialect)
    normalized_value = normalized_expression.sql(dialect=to_dialect, pretty=pretty)
    if isinstance(value, str) and is_pattern:
        for pos in star_positions:
            normalized_value = normalized_value[:pos] + "*" + normalized_value[pos:]
    return normalized_value



class SQLFrameException(Exception):
    pass


class UnsupportedOperationError(SQLFrameException):
    pass


class RowError(SQLFrameException):
    pass


class TableSchemaError(SQLFrameException):
    pass


class PandasDiffError(SQLFrameException):
    pass


class DataFrameDiffError(SQLFrameException):
    pass


class SchemaDiffError(SQLFrameException):
    pass



from __future__ import annotations

import sys
import time

import pytest


def pytest_collection_modifyitems(items, *args, **kwargs):
    for item in items:
        if not [x for x in item.iter_markers() if x.name != "parametrize"]:
            item.add_marker("fast")


@pytest.fixture(scope="session", autouse=True)
def set_tz():
    import os

    os.environ["TZ"] = "UTC"
    time.tzset()
    yield
    del os.environ["TZ"]


@pytest.fixture(scope="function", autouse=True)
def rescope_sparksession_singleton():
    from sqlframe.base.session import _BaseSession

    _BaseSession._instance = None
    yield



from __future__ import annotations

import logging
import typing as t

from sqlframe.base.catalog import Column as CatalogColumn
from sqlframe.base.dataframe import (
    _BaseDataFrame,
    _BaseDataFrameNaFunctions,
    _BaseDataFrameStatFunctions,
)
from sqlframe.base.mixins.dataframe_mixins import NoCachePersistSupportMixin
from sqlframe.base.util import normalize_string
from sqlframe.databricks.group import DatabricksGroupedData

if t.TYPE_CHECKING:
    from sqlframe.databricks.readwriter import DatabricksDataFrameWriter
    from sqlframe.databricks.session import DatabricksSession


logger = logging.getLogger(__name__)


class DatabricksDataFrameNaFunctions(_BaseDataFrameNaFunctions["DatabricksDataFrame"]):
    pass


class DatabricksDataFrameStatFunctions(_BaseDataFrameStatFunctions["DatabricksDataFrame"]):
    pass


class DatabricksDataFrame(
    NoCachePersistSupportMixin,
    _BaseDataFrame[
        "DatabricksSession",
        "DatabricksDataFrameWriter",
        "DatabricksDataFrameNaFunctions",
        "DatabricksDataFrameStatFunctions",
        "DatabricksGroupedData",
    ],
):
    _na = DatabricksDataFrameNaFunctions
    _stat = DatabricksDataFrameStatFunctions
    _group_data = DatabricksGroupedData

    @property
    def _typed_columns(self) -> t.List[CatalogColumn]:
        sql = self.session._to_sql(self.expression)
        columns = []
        for row in self.session._collect(r"DESCRIBE QUERY ({sql})".format(sql=sql)):
            columns.append(
                CatalogColumn(
                    name=normalize_string(
                        row.col_name,
                        from_dialect="execution",
                        to_dialect="output",
                    ),
                    dataType=normalize_string(
                        row.data_type,
                        from_dialect="execution",
                        to_dialect="output",
                        is_datatype=True,
                    ),
                    nullable=True,
                    description=row.comment,
                    isPartition=False,
                    isBucket=False,
                )
            )
        return columns



from __future__ import annotations

import inspect
import sys

import sqlframe.base.functions  # noqa

module = sys.modules["sqlframe.base.functions"]
globals().update(
    {
        name: func
        for name, func in inspect.getmembers(module, inspect.isfunction)
        if hasattr(func, "unsupported_engines")
        and "duckdb" not in func.unsupported_engines
        and "*" not in func.unsupported_engines
    }
)


from sqlframe.base.function_alternatives import (  # noqa
    any_value_always_ignore_nulls as any_value,
    e_literal as e,
    expm1_from_exp as expm1,
    log1p_from_log as log1p,
    rint_from_round as rint,
    kurtosis_from_kurtosis_pop as kurtosis,
    collect_set_from_list_distinct as collect_set,
    first_always_ignore_nulls as first,
    factorial_ensure_int as factorial,
    isnull_using_equal as isnull,
    nanvl_as_case as nanvl,
    percentile_approx_without_accuracy as percentile_approx,
    rand_no_seed as rand,
    base64_from_blob as base64,
    decode_from_blob as decode,
    format_string_with_pipes as format_string,
    overlay_from_substr as overlay,
    split_no_limit as split,
    arrays_overlap_using_intersect as arrays_overlap,
    slice_as_list_slice as slice,
    array_join_null_replacement_with_transform as array_join,
    element_at_using_brackets as element_at,
    array_remove_using_filter as array_remove,
    array_union_using_list_concat as array_union,
    array_min_from_sort as array_min,
    array_max_from_sort as array_max,
    sequence_from_generate_series as sequence,
    try_element_at_zero_based as try_element_at,
    day_with_try_to_timestamp as day,
    try_to_timestamp_strptime as try_to_timestamp,
    _is_string_using_typeof_varchar as _is_string,
    array_append_list_append as array_append,
    endswith_with_underscore as endswith,
    last_day_with_cast as last_day,
    regexp_replace_global_option as regexp_replace,
    to_timestamp_tz as to_timestamp,
)



from __future__ import annotations

import typing as t

import pytest
from _pytest.fixtures import FixtureRequest
from pyspark.sql import DataFrame as PySparkDataFrame
from pyspark.sql import functions as F

from sqlframe.standalone import functions as SF
from sqlframe.standalone.dataframe import StandaloneDataFrame
from tests.integration.fixtures import StandaloneSession

if t.TYPE_CHECKING:
    from sqlframe.base.dataframe import _BaseDataFrame

    DataFrameMapping = t.Dict[str, _BaseDataFrame]

pytest_plugins = ["tests.integration.fixtures"]


def test_empty_df(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    df_empty = pyspark_employee.sparkSession.createDataFrame([], "cola int, colb int")
    dfs_empty = get_df("employee").session.createDataFrame([], "cola int, colb int")
    compare_frames(df_empty, dfs_empty, no_empty=False)


def test_simple_select(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.select(F.col("employee_id"))
    dfs_employee = employee.select(SF.col("employee_id"))
    compare_frames(df_employee, dfs_employee)


def test_simple_select_from_table(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee
    dfs = employee.session.read.table("employee")
    compare_frames(df, dfs)


def test_select_star_from_table(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    df = pyspark_employee
    dfs = get_df("employee").session.read.table("employee")
    compare_frames(df, dfs)


def test_simple_select_df_attribute(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.select(pyspark_employee.employee_id)
    dfs_employee = employee.select(employee.employee_id)
    compare_frames(df_employee, dfs_employee)


def test_simple_select_df_dict(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.select(pyspark_employee["employee_id"])
    dfs_employee = employee.select(employee["employee_id"])
    compare_frames(df_employee, dfs_employee)


def test_multiple_selects(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.select(
        pyspark_employee["employee_id"], F.col("fname"), pyspark_employee.lname
    )
    dfs_employee = employee.select(employee["employee_id"], SF.col("fname"), employee.lname)
    compare_frames(df_employee, dfs_employee)


def test_alias_no_op(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.alias("df_employee")
    dfs_employee = employee.alias("dfs_employee")
    compare_frames(df_employee, dfs_employee)


def test_alias_with_select(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.alias("df_employee").select(
        pyspark_employee["employee_id"],
        F.col("df_employee.fname"),
        pyspark_employee.lname,
    )
    dfs_employee = employee.alias("dfs_employee").select(
        employee["employee_id"],
        SF.col("dfs_employee.fname"),
        employee.lname,
    )
    compare_frames(df_employee, dfs_employee)


def test_alias_with_space(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = (
        pyspark_employee.alias("the employees")
        .select(F.col("the employees.fname").alias("first name"))
        .limit(100)
        .select(F.col("first name").alias("blah blah"))
    )
    dfs_employee = (
        employee.alias("the employees")
        .select(SF.col("the employees.fname").alias("first name"))
        .limit(100)
        .select(SF.col("first name").alias("blah blah"))
    )
    compare_frames(df_employee, dfs_employee)


def test_case_when_otherwise(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(
            (F.col("age") >= F.lit(40)) & (F.col("age") <= F.lit(60)),
            F.lit("between 40 and 60"),
        )
        .when(F.col("age") < F.lit(40), "less than 40")
        .otherwise("greater than 60")
    )

    dfs = employee.select(
        SF.when(
            (SF.col("age") >= SF.lit(40)) & (SF.col("age") <= SF.lit(60)),
            SF.lit("between 40 and 60"),
        )
        .when(SF.col("age") < SF.lit(40), "less than 40")
        .otherwise("greater than 60")
    )

    compare_frames(df, dfs, compare_schema=False)


def test_case_when_no_otherwise(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(
            (F.col("age") >= F.lit(40)) & (F.col("age") <= F.lit(60)),
            F.lit("between 40 and 60"),
        ).when(F.col("age") < F.lit(40), "less than 40")
    )

    dfs = employee.select(
        SF.when(
            (SF.col("age") >= SF.lit(40)) & (SF.col("age") <= SF.lit(60)),
            SF.lit("between 40 and 60"),
        ).when(SF.col("age") < SF.lit(40), "less than 40")
    )

    compare_frames(df, dfs, compare_schema=False)


def test_case_when_implicit_lit(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(
            F.col("fname") == "Jack",
            "name is Jack",
        ).otherwise("Default")
    )

    dfs = employee.select(
        SF.when(
            (SF.col("fname") == "Jack"),
            "name is Jack",
        ).otherwise("Default")
    )

    compare_frames(df, dfs, compare_schema=False)


def test_where_clause_single(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(F.col("age") == F.lit(37))
    dfs_employee = employee.where(SF.col("age") == SF.lit(37))
    compare_frames(df_employee, dfs_employee)


def test_where_clause_eq_nullsafe(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(F.col("age").eqNullSafe(F.lit(37)))
    dfs_employee = employee.where(SF.col("age") == SF.lit(37))
    compare_frames(df_employee, dfs_employee)


def test_where_clause_multiple_and(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(
        (F.col("age") == F.lit(37)) & (F.col("fname") == F.lit("Jack"))
    )
    dfs_employee = employee.where(
        (SF.col("age") == SF.lit(37)) & (SF.col("fname") == SF.lit("Jack"))
    )
    compare_frames(df_employee, dfs_employee)


def test_where_many_and(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(
        (F.col("age") == F.lit(37))
        & (F.col("fname") == F.lit("Jack"))
        & (F.col("lname") == F.lit("Shephard"))
        & (F.col("employee_id") == F.lit(1))
    )
    dfs_employee = employee.where(
        (SF.col("age") == SF.lit(37))
        & (SF.col("fname") == SF.lit("Jack"))
        & (SF.col("lname") == SF.lit("Shephard"))
        & (SF.col("employee_id") == SF.lit(1))
    )
    compare_frames(df_employee, dfs_employee)


def test_where_clause_multiple_or(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(
        (F.col("age") == F.lit(37)) | (F.col("fname") == F.lit("Kate"))
    )
    dfs_employee = employee.where(
        (SF.col("age") == SF.lit(37)) | (SF.col("fname") == SF.lit("Kate"))
    )
    compare_frames(df_employee, dfs_employee)


def test_where_many_or(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(
        (F.col("age") == F.lit(37))
        | (F.col("fname") == F.lit("Kate"))
        | (F.col("lname") == F.lit("Littleton"))
        | (F.col("employee_id") == F.lit(2))
    )
    dfs_employee = employee.where(
        (SF.col("age") == SF.lit(37))
        | (SF.col("fname") == SF.lit("Kate"))
        | (SF.col("lname") == SF.lit("Littleton"))
        | (SF.col("employee_id") == SF.lit(2))
    )
    compare_frames(df_employee, dfs_employee)


def test_where_mixed_and_or(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(
        ((F.col("age") == F.lit(65)) & (F.col("fname") == F.lit("John")))
        | ((F.col("lname") == F.lit("Shephard")) & (F.col("age") == F.lit(37)))
    )
    dfs_employee = employee.where(
        ((SF.col("age") == SF.lit(65)) & (SF.col("fname") == SF.lit("John")))
        | ((SF.col("lname") == SF.lit("Shephard")) & (SF.col("age") == SF.lit(37)))
    )
    compare_frames(df_employee, dfs_employee)


def test_where_multiple_chained(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(F.col("age") == F.lit(37)).where(
        pyspark_employee.fname == F.lit("Jack")
    )
    dfs_employee = employee.where(SF.col("age") == SF.lit(37)).where(
        employee.fname == SF.lit("Jack")
    )
    compare_frames(df_employee, dfs_employee)


def test_where_sql_expr(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where("age = 37 AND fname = 'Jack'")
    dfs_employee = employee.where("age = 37 AND fname = 'Jack'")
    compare_frames(df_employee, dfs_employee)


def test_operators(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_employee = pyspark_employee.where(pyspark_employee["age"] < F.lit(50))
    dfs_employee = employee.where(employee["age"] < SF.lit(50))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] <= F.lit(37))
    dfs_employee = employee.where(employee["age"] <= SF.lit(37))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] > F.lit(50))
    dfs_employee = employee.where(employee["age"] > SF.lit(50))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] >= F.lit(37))
    dfs_employee = employee.where(employee["age"] >= SF.lit(37))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] != F.lit(50))
    dfs_employee = employee.where(employee["age"] != SF.lit(50))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] == F.lit(37))
    dfs_employee = employee.where(employee["age"] == SF.lit(37))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] % F.lit(5) == F.lit(0))
    dfs_employee = employee.where(employee["age"] % SF.lit(5) == SF.lit(0))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] + F.lit(5) > F.lit(28))
    dfs_employee = employee.where(employee["age"] + SF.lit(5) > SF.lit(28))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(pyspark_employee["age"] - F.lit(5) > F.lit(28))
    dfs_employee = employee.where(employee["age"] - SF.lit(5) > SF.lit(28))
    compare_frames(df_employee, dfs_employee)

    df_employee = pyspark_employee.where(
        pyspark_employee["age"] * F.lit(0.5) == pyspark_employee["age"] / F.lit(2)
    )
    dfs_employee = employee.where(employee["age"] * SF.lit(0.5) == employee["age"] / SF.lit(2))
    compare_frames(df_employee, dfs_employee)


def test_join_inner(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(pyspark_store, on=["store_id"], how="inner").select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        F.coalesce("store_id", "age").alias("store_id_size"),
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(store, on=["store_id"], how="inner").select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        SF.coalesce("store_id", "age").alias("store_id_size"),
        store.store_name,
        store["num_sales"],
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_join_inner_no_select(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.select(F.col("store_id"), F.col("fname"), F.col("lname")).join(
        pyspark_store.select(F.col("store_id"), F.col("store_name")),
        on=["store_id"],
        how="inner",
    )
    dfs_joined = employee.select(SF.col("store_id"), SF.col("fname"), SF.col("lname")).join(
        store.select(SF.col("store_id"), SF.col("store_name")),
        on=["store_id"],
        how="inner",
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_join_inner_equality_single(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        pyspark_store,
        on=pyspark_employee.store_id == pyspark_store.store_id,
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
        F.lit("literal_value"),
    )
    dfs_joined = employee.join(
        store,
        on=employee.store_id == store.store_id,
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
        SF.lit("literal_value"),
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_join_inner_equality_multiple(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        pyspark_store,
        on=[
            pyspark_employee.store_id == pyspark_store.store_id,
            pyspark_employee.age == pyspark_store.num_sales,
        ],
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(
        store,
        on=[
            employee.store_id == store.store_id,
            employee.age == store.num_sales,
        ],
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_join_inner_equality_multiple_bitwise_and(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        pyspark_store,
        on=(pyspark_store.store_id == pyspark_employee.store_id)
        & (pyspark_store.num_sales == pyspark_employee.age),
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(
        store,
        on=(store.store_id == employee.store_id) & (store.num_sales == employee.age),
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_join_left_outer(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = (
        pyspark_employee.join(pyspark_store, on=["store_id"], how="left_outer")
        .select(
            pyspark_employee.employee_id,
            pyspark_employee["fname"],
            F.col("lname"),
            F.col("age"),
            F.col("store_id"),
            pyspark_store.store_name,
            pyspark_store["num_sales"],
        )
        .orderBy(F.col("employee_id"))
    )
    dfs_joined = (
        employee.join(store, on=["store_id"], how="left_outer")
        .select(
            employee.employee_id,
            employee["fname"],
            SF.col("lname"),
            SF.col("age"),
            SF.col("store_id"),
            store.store_name,
            store["num_sales"],
        )
        .orderBy(SF.col("employee_id"))
    )
    compare_frames(df_joined, dfs_joined)


def test_join_full_outer(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_bigquery: t.Callable,
):
    if is_bigquery():
        pytest.skip(
            "BigQuery doesn't support full outer joins on unnested arrays. You get `Array scan is not allowed with FULL JOIN`"
        )
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(pyspark_store, on=["store_id"], how="full_outer").select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        F.col("store_id"),
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(store, on=["store_id"], how="full_outer").select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        SF.col("store_id"),
        store.store_name,
        store["num_sales"],
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_triple_join(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df = (
        pyspark_employee.join(
            pyspark_store, on=pyspark_employee.employee_id == pyspark_store.store_id
        )
        .join(pyspark_district, on=pyspark_store.store_id == pyspark_district.district_id)
        .select(
            pyspark_employee.employee_id,
            pyspark_store.store_id,
            pyspark_district.district_id,
            pyspark_employee.fname,
            pyspark_store.store_name,
            pyspark_district.district_name,
        )
    )
    dfs = (
        employee.join(store, on=employee.employee_id == store.store_id)
        .join(district, on=store.store_id == district.district_id)
        .select(
            employee.employee_id,
            store.store_id,
            district.district_id,
            employee.fname,
            store.store_name,
            district.district_name,
        )
    )
    compare_frames(df, dfs, sort=True)


def test_triple_join_no_select(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_duckdb: t.Callable,
    is_postgres: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
    is_spark: t.Callable,
):
    if is_duckdb():
        pytest.skip(
            "Duckdb doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_postgres():
        pytest.skip(
            "Postgres doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_redshift():
        pytest.skip(
            "Redshift doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_snowflake():
        pytest.skip(
            "Snowflake doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_spark():
        pytest.skip("Spark doesn't support duplicate column names")
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df = (
        pyspark_employee.join(
            pyspark_store,
            on=pyspark_employee["employee_id"] == pyspark_store["store_id"],
            how="left",
        )
        .join(
            pyspark_district,
            on=pyspark_store["store_id"] == pyspark_district["district_id"],
            how="left",
        )
        .orderBy(F.col("employee_id"))
    )
    dfs = (
        employee.join(
            store,
            on=employee["employee_id"] == store["store_id"],
            how="left",
        )
        .join(
            district,
            on=store["store_id"] == district["district_id"],
            how="left",
        )
        .orderBy(SF.col("employee_id"))
    )
    compare_frames(df, dfs, sort=True)


def test_triple_joins_filter(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_duckdb: t.Callable,
    is_postgres: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
    is_spark: t.Callable,
):
    if is_duckdb():
        pytest.skip(
            "Duckdb doesn't support duplicate column names and they will just be reduced to a single column where the last one wins"
        )
    if is_postgres():
        pytest.skip(
            "Postgres doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_redshift():
        pytest.skip(
            "Redshift doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_snowflake():
        pytest.skip(
            "Snowflake doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_spark():
        pytest.skip("Spark doesn't support duplicate column names")
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df = (
        pyspark_employee.join(
            pyspark_store,
            on=pyspark_employee["employee_id"] == pyspark_store["store_id"],
            how="left",
        ).join(
            pyspark_district,
            on=pyspark_store["store_id"] == pyspark_district["district_id"],
            how="left",
        )
    ).filter(F.coalesce(pyspark_store["num_sales"], F.lit(0)) > 100)
    dfs = (
        employee.join(
            store,
            on=employee["employee_id"] == store["store_id"],
            how="left",
        ).join(
            district,
            on=store["store_id"] == district["district_id"],
            how="left",
        )
    ).filter(SF.coalesce(store["num_sales"], SF.lit(0)) > 100)
    compare_frames(df, dfs, sort=True)


def test_triple_join_column_name_only(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_duckdb: t.Callable,
    is_postgres: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
    is_spark: t.Callable,
):
    if is_duckdb():
        pytest.skip(
            "Duckdb doesn't support duplicate column names and they will just be reduced to a single column where the last one wins"
        )
    if is_postgres():
        pytest.skip(
            "Postgres doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_redshift():
        pytest.skip(
            "Redshift doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_snowflake():
        pytest.skip(
            "Snowflake doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_spark():
        pytest.skip("Spark doesn't support duplicate column names")
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df = (
        pyspark_employee.join(
            pyspark_store,
            on=pyspark_employee["employee_id"] == pyspark_store["store_id"],
            how="left",
        )
        .join(pyspark_district, on="district_id", how="left")
        .orderBy(F.col("employee_id"))
    )
    dfs = (
        employee.join(
            store,
            on=employee["employee_id"] == store["store_id"],
            how="left",
        )
        .join(district, on="district_id", how="left")
        .orderBy(SF.col("employee_id"))
    )
    compare_frames(df, dfs, sort=True)


def test_join_select_and_select_start(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df = pyspark_employee.select(
        F.col("fname"), F.col("lname"), F.col("age"), F.col("store_id")
    ).join(pyspark_store, "store_id", "inner")

    dfs = employee.select(SF.col("fname"), SF.col("lname"), SF.col("age"), SF.col("store_id")).join(
        store, "store_id", "inner"
    )

    compare_frames(df, dfs, sort=True)


def test_join_no_on(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    # No on results in a cross. Testing that "how" is ignored
    pyspark_employee = pyspark_employee.select(F.col("fname"), F.col("lname"))
    pyspark_store = pyspark_store.select(F.col("store_id"), F.col("num_sales"))
    employee = get_df("employee").select(SF.col("fname"), SF.col("lname"))
    store = get_df("store").select(SF.col("store_id"), SF.col("num_sales"))
    df = pyspark_employee.join(pyspark_store, how="inner")
    dfs = employee.join(store, how="inner")

    compare_frames(df, dfs)


def test_cross_join(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    # No on results in a cross. Testing that "how" is ignored
    pyspark_employee = pyspark_employee.select(F.col("fname"), F.col("lname"))
    pyspark_store = pyspark_store.select(F.col("store_id"), F.col("num_sales"))
    employee = get_df("employee").select(SF.col("fname"), SF.col("lname"))
    store = get_df("store").select(SF.col("store_id"), SF.col("num_sales"))
    df = pyspark_employee.crossJoin(pyspark_store)
    dfs = employee.crossJoin(store)

    compare_frames(df, dfs)


def test_branching_root_dataframes(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_duckdb: t.Callable,
    is_postgres: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
    is_spark: t.Callable,
):
    """
    Test a pattern that has non-intuitive behavior in spark

    Scenario: You do a self-join in a dataframe using an original dataframe and then a modified version
    of it. You then reference the columns by the dataframe name instead of the column function.
    Spark will use the root dataframe's column in the result.
    """
    if is_duckdb():
        pytest.skip(
            "Duckdb doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_postgres():
        pytest.skip(
            "Postgres doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_redshift():
        pytest.skip(
            "Redshift doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_snowflake():
        pytest.skip(
            "Snowflake doesn't support duplicate column names and they will just be reduced to a single column"
        )
    if is_spark():
        pytest.skip(
            "Spark doesn't support duplicate column names and they will just be reduced to a single column."
        )
    employee = get_df("employee")
    df_hydra_employees_only = pyspark_employee.where(F.col("store_id") == F.lit(1))
    df_joined = (
        pyspark_employee.where(F.col("store_id") == F.lit(2))
        .alias("df_arrow_employees_only")
        .join(
            df_hydra_employees_only.alias("df_hydra_employees_only"),
            on=["store_id"],
            how="full_outer",
        )
        .select(
            pyspark_employee.fname,
            F.col("df_arrow_employees_only.fname"),
            df_hydra_employees_only.fname,
            F.col("df_hydra_employees_only.fname"),
        )
    )

    dfs_hydra_employees_only = employee.where(SF.col("store_id") == SF.lit(1))
    dfs_joined = (
        employee.where(SF.col("store_id") == SF.lit(2))
        .alias("dfs_arrow_employees_only")
        .join(
            dfs_hydra_employees_only.alias("dfs_hydra_employees_only"),
            on=["store_id"],
            how="full_outer",
        )
        .select(
            employee.fname,
            SF.col("dfs_arrow_employees_only.fname"),
            dfs_hydra_employees_only.fname,
            SF.col("dfs_hydra_employees_only.fname"),
        )
    )
    compare_frames(df_joined, dfs_joined, sort=True)


def test_basic_union(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_unioned = pyspark_employee.select(F.col("employee_id"), F.col("age")).union(
        pyspark_store.select(F.col("store_id"), F.col("num_sales"))
    )

    dfs_unioned = employee.select(SF.col("employee_id"), SF.col("age")).union(
        store.select(SF.col("store_id"), SF.col("num_sales"))
    )
    compare_frames(df_unioned, dfs_unioned)


def test_union_with_join(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df_joined = pyspark_employee.join(
        pyspark_store,
        on="store_id",
        how="inner",
    )
    df_unioned = df_joined.select(F.col("store_id"), F.col("store_name")).union(
        pyspark_district.select(F.col("district_id"), F.col("district_name"))
    )

    dfs_joined = employee.join(
        store,
        on="store_id",
        how="inner",
    )
    dfs_unioned = dfs_joined.select(SF.col("store_id"), SF.col("store_name")).union(
        district.select(SF.col("district_id"), SF.col("district_name"))
    )

    compare_frames(df_unioned, dfs_unioned)


def test_double_union_all(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")
    df_unioned = (
        pyspark_employee.select(F.col("employee_id"), F.col("fname"))
        .unionAll(pyspark_store.select(F.col("store_id"), F.col("store_name")))
        .unionAll(pyspark_district.select(F.col("district_id"), F.col("district_name")))
    )

    dfs_unioned = (
        employee.select(SF.col("employee_id"), SF.col("fname"))
        .unionAll(store.select(SF.col("store_id"), SF.col("store_name")))
        .unionAll(district.select(SF.col("district_id"), SF.col("district_name")))
    )

    compare_frames(df_unioned, dfs_unioned)


def test_union_by_name(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df = pyspark_employee.select(F.col("employee_id"), F.col("fname"), F.col("lname")).unionByName(
        pyspark_store.select(
            F.col("store_name").alias("lname"),
            F.col("store_id").alias("employee_id"),
            F.col("store_name").alias("fname"),
        )
    )

    dfs = employee.select(SF.col("employee_id"), SF.col("fname"), SF.col("lname")).unionByName(
        store.select(
            SF.col("store_name").alias("lname"),
            SF.col("store_id").alias("employee_id"),
            SF.col("store_name").alias("fname"),
        )
    )

    compare_frames(df, dfs)


def test_union_by_name_allow_missing(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_postgres: t.Callable,
):
    if is_postgres():
        pytest.skip(
            "For columns that are missing they are replaced with null and then postgres errors complaining about unioning null with another type"
        )
    employee = get_df("employee")
    store = get_df("store")
    df = pyspark_employee.select(
        F.col("age"), F.col("employee_id"), F.col("fname"), F.col("lname")
    ).unionByName(
        pyspark_store.select(
            F.col("store_name").alias("lname"),
            F.col("store_id").alias("employee_id"),
            F.col("store_name").alias("fname"),
            F.col("num_sales"),
        ),
        allowMissingColumns=True,
    )

    dfs = employee.select(
        SF.col("age"), SF.col("employee_id"), SF.col("fname"), SF.col("lname")
    ).unionByName(
        store.select(
            SF.col("store_name").alias("lname"),
            SF.col("store_id").alias("employee_id"),
            SF.col("store_name").alias("fname"),
            SF.col("num_sales"),
        ),
        allowMissingColumns=True,
    )

    compare_frames(df, dfs)


def test_order_by_default(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales"))
        .orderBy(F.col("district_id"))
    )

    dfs = (
        store.groupBy(SF.col("district_id")).agg(SF.min("num_sales")).orderBy(SF.col("district_id"))
    )

    compare_frames(df, dfs, compare_schema=False)


def test_order_by_array_bool(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales").alias("total_sales"))
        .orderBy(F.col("total_sales"), F.col("district_id"), ascending=[1, 0])
    )

    dfs = (
        store.groupBy(SF.col("district_id"))
        .agg(SF.min("num_sales").alias("total_sales"))
        .orderBy(SF.col("total_sales"), SF.col("district_id"), ascending=[1, 0])
    )

    compare_frames(df, dfs)


def test_order_by_single_bool(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales").alias("total_sales"))
        .orderBy(F.col("total_sales"), F.col("district_id"), ascending=False)
    )

    dfs = (
        store.groupBy(SF.col("district_id"))
        .agg(SF.min("num_sales").alias("total_sales"))
        .orderBy(SF.col("total_sales"), SF.col("district_id"), ascending=False)
    )

    compare_frames(df, dfs)


def test_order_by_column_sort_method(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales").alias("total_sales"))
        .orderBy(F.col("total_sales").asc(), F.col("district_id").desc())
    )

    dfs = (
        store.groupBy(SF.col("district_id"))
        .agg(SF.min("num_sales").alias("total_sales"))
        .orderBy(SF.col("total_sales").asc(), SF.col("district_id").desc())
    )

    compare_frames(df, dfs)


def test_order_by_column_sort_method_nulls_last(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales").alias("total_sales"))
        .orderBy(F.when(F.col("district_id") == F.lit(2), F.col("district_id")).asc_nulls_last())
    )

    dfs = (
        store.groupBy(SF.col("district_id"))
        .agg(SF.min("num_sales").alias("total_sales"))
        .orderBy(
            SF.when(SF.col("district_id") == SF.lit(2), SF.col("district_id")).asc_nulls_last()
        )
    )

    compare_frames(df, dfs)


def test_order_by_column_sort_method_nulls_first(
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    store = get_df("store")
    df = (
        pyspark_store.groupBy(F.col("district_id"))
        .agg(F.min("num_sales").alias("total_sales"))
        .orderBy(F.when(F.col("district_id") == F.lit(1), F.col("district_id")).desc_nulls_first())
    )

    dfs = (
        store.groupBy(SF.col("district_id"))
        .agg(SF.min("num_sales").alias("total_sales"))
        .orderBy(
            SF.when(SF.col("district_id") == SF.lit(1), SF.col("district_id")).desc_nulls_first()
        )
    )

    compare_frames(df, dfs)


def test_intersect(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_employee_duplicate = pyspark_employee.select(F.col("employee_id"), F.col("store_id")).union(
        pyspark_employee.select(F.col("employee_id"), F.col("store_id"))
    )

    df_store_duplicate = pyspark_store.select(F.col("store_id"), F.col("district_id")).union(
        pyspark_store.select(F.col("store_id"), F.col("district_id"))
    )

    df = df_employee_duplicate.intersect(df_store_duplicate)

    dfs_employee_duplicate = employee.select(SF.col("employee_id"), SF.col("store_id")).union(
        employee.select(SF.col("employee_id"), SF.col("store_id"))
    )

    dfs_store_duplicate = store.select(SF.col("store_id"), SF.col("district_id")).union(
        store.select(SF.col("store_id"), SF.col("district_id"))
    )

    dfs = dfs_employee_duplicate.intersect(dfs_store_duplicate)

    compare_frames(df, dfs)


def test_intersect_all(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_bigquery: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
):
    if is_bigquery():
        pytest.skip("BigQuery doesn't support INTERSECT ALL. Only INTERSECT DISTINCT is supported")
    if is_redshift():
        pytest.skip("Redshift doesn't support INTERSECT ALL. Only INTERSECT DISTINCT is supported")
    if is_snowflake():
        pytest.skip("Snowflake doesn't support INTERSECT ALL. Only INTERSECT DISTINCT is supported")
    employee = get_df("employee")
    store = get_df("store")
    df_employee_duplicate = pyspark_employee.select(F.col("employee_id"), F.col("store_id")).union(
        pyspark_employee.select(F.col("employee_id"), F.col("store_id"))
    )

    df_store_duplicate = pyspark_store.select(F.col("store_id"), F.col("district_id")).union(
        pyspark_store.select(F.col("store_id"), F.col("district_id"))
    )

    df = df_employee_duplicate.intersectAll(df_store_duplicate)

    dfs_employee_duplicate = employee.select(SF.col("employee_id"), SF.col("store_id")).union(
        employee.select(SF.col("employee_id"), SF.col("store_id"))
    )

    dfs_store_duplicate = store.select(SF.col("store_id"), SF.col("district_id")).union(
        store.select(SF.col("store_id"), SF.col("district_id"))
    )

    dfs = dfs_employee_duplicate.intersectAll(dfs_store_duplicate)

    compare_frames(df, dfs)


def test_except_all(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_bigquery: t.Callable,
    is_redshift: t.Callable,
    is_snowflake: t.Callable,
):
    if is_bigquery():
        pytest.skip("BigQuery doesn't support EXCEPT ALL. Only EXCEPT DISTINCT is supported")
    if is_redshift():
        pytest.skip("Redshift doesn't support EXCEPT ALL. Only EXCEPT DISTINCT is supported")
    if is_snowflake():
        pytest.skip("Snowflake doesn't support EXCEPT ALL. Only EXCEPT DISTINCT is supported")
    employee = get_df("employee")
    store = get_df("store")
    df_employee_duplicate = pyspark_employee.select(F.col("employee_id"), F.col("store_id")).union(
        pyspark_employee.select(F.col("employee_id"), F.col("store_id"))
    )

    df_store_duplicate = pyspark_store.select(F.col("store_id"), F.col("district_id")).union(
        pyspark_store.select(F.col("store_id"), F.col("district_id"))
    )

    df = df_employee_duplicate.exceptAll(df_store_duplicate)

    dfs_employee_duplicate = employee.select(SF.col("employee_id"), SF.col("store_id")).union(
        employee.select(SF.col("employee_id"), SF.col("store_id"))
    )

    dfs_store_duplicate = store.select(SF.col("store_id"), SF.col("district_id")).union(
        store.select(SF.col("store_id"), SF.col("district_id"))
    )

    dfs = dfs_employee_duplicate.exceptAll(dfs_store_duplicate)

    compare_frames(df, dfs)


def test_distinct(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("age")).distinct()

    dfs = employee.select(SF.col("age")).distinct()

    compare_frames(df, dfs)


def test_union_distinct(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df_unioned = (
        pyspark_employee.select(F.col("employee_id"), F.col("age"))
        .union(pyspark_employee.select(F.col("employee_id"), F.col("age")))
        .distinct()
    )

    dfs_unioned = (
        employee.select(SF.col("employee_id"), SF.col("age"))
        .union(employee.select(SF.col("employee_id"), SF.col("age")))
        .distinct()
    )
    compare_frames(df_unioned, dfs_unioned)


def test_drop_duplicates_no_subset(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select("age").dropDuplicates()
    dfs = employee.select("age").dropDuplicates()
    compare_frames(df, dfs)


def test_drop_duplicates_subset(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_redshift: t.Callable,
):
    if is_redshift():
        # The expected behavior for drop duplicates with subset is not very clear since there is no explicit ordering
        # As a result Redshift drops the first duplicate. Could consider changing this behavior but this function
        # has general determinism issue so just skipping for now.
        pytest.skip("Redshift doesn't match the expected behavior for drop_duplicates with subset")
    employee = get_df("employee")
    df = pyspark_employee.dropDuplicates(["age"])
    dfs = employee.dropDuplicates(["age"])
    compare_frames(df, dfs)


def test_drop_na_default(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age")
    ).dropna()

    dfs = employee.select(
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age")
    ).dropna()

    compare_frames(df, dfs)


def test_dropna_how(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.lit(None), F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age")
    ).dropna(how="all")

    dfs = employee.select(
        SF.lit(None), SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age")
    ).dropna(how="all")

    compare_frames(df, dfs, compare_schema=False)


def test_dropna_thresh(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.lit(None).alias("val"),
        F.lit(1),
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age"),
    ).dropna(how="any", thresh=2)

    dfs = employee.select(
        SF.lit(None).alias("val"),
        SF.lit(1),
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age"),
    ).dropna(how="any", thresh=2)

    compare_frames(df, dfs, compare_schema=False)


def test_dropna_subset(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.lit(None).alias("val"),
        F.lit(1),
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age"),
    ).dropna(thresh=1, subset="the_age")

    dfs = employee.select(
        SF.lit(None).alias("val"),
        SF.lit(1),
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age"),
    ).dropna(thresh=1, subset="the_age")

    compare_frames(df, dfs, compare_schema=False)


def test_dropna_na_function(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age")
    ).na.drop()

    dfs = employee.select(
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age")
    ).na.drop()

    compare_frames(df, dfs)


def test_fillna_default(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age")
    ).fillna(100)

    dfs = employee.select(
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age")
    ).fillna(100)

    compare_frames(df, dfs)


def test_fillna_dict_replacement(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.col("fname"),
        F.when(F.col("lname").startswith("L"), F.col("lname")).alias("l_lname"),
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age"),
    ).fillna({"fname": "Jacob", "l_lname": "NOT_LNAME"})

    dfs = employee.select(
        SF.col("fname"),
        SF.when(SF.col("lname").startswith("L"), SF.col("lname")).alias("l_lname"),
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age"),
    ).fillna({"fname": "Jacob", "l_lname": "NOT_LNAME"})

    # For some reason the sqlglot results sets a column as nullable when it doesn't need to
    # This seems to be a nuance in how spark dataframe from sql works so we can ignore
    compare_frames(df, dfs, compare_schema=False)


def test_fillna_na_func(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.when(F.col("age") < F.lit(50), F.col("age")).alias("the_age")
    ).na.fill(100)

    dfs = employee.select(
        SF.when(SF.col("age") < SF.lit(50), SF.col("age")).alias("the_age")
    ).na.fill(100)

    compare_frames(df, dfs)


def test_replace_basic(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("age"), F.lit(37).alias("test_col")).replace(
        to_replace=37, value=100
    )

    dfs = employee.select(SF.col("age"), SF.lit(37).alias("test_col")).replace(
        to_replace=37, value=100
    )
    compare_frames(df, dfs, compare_schema=False)


def test_replace_basic_subset(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("age"), F.lit(37).alias("test_col")).replace(  # type: ignore
        to_replace=37, value=100, subset="age"
    )

    dfs = employee.select(SF.col("age"), SF.lit(37).alias("test_col")).replace(
        to_replace=37, value=100, subset="age"
    )

    compare_frames(df, dfs, compare_schema=False)


def test_replace_mapping(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("age"), F.lit(37).alias("test_col")).replace({37: 100})

    dfs = employee.select(SF.col("age"), SF.lit(37).alias("test_col")).replace({37: 100})

    compare_frames(df, dfs, compare_schema=False)


def test_replace_mapping_subset(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(
        F.col("age"), F.lit(37).alias("test_col"), F.lit(50).alias("test_col_2")
    ).replace({37: 100, 50: 1}, subset=["age", "test_col_2"])

    dfs = employee.select(
        SF.col("age"), SF.lit(37).alias("test_col"), SF.lit(50).alias("test_col_2")
    ).replace({37: 100, 50: 1}, subset=["age", "test_col_2"])

    compare_frames(df, dfs, compare_schema=False)


def test_replace_na_func_basic(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("age"), F.lit(37).alias("test_col")).na.replace(  # type: ignore
        to_replace=37, value=100
    )

    dfs = employee.select(SF.col("age"), SF.lit(37).alias("test_col")).na.replace(
        to_replace=37, value=100
    )

    compare_frames(df, dfs, compare_schema=False)


def test_with_column(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.withColumn("test", F.col("age"))

    dfs = employee.withColumn("test", SF.col("age"))

    compare_frames(df, dfs)


def test_with_column_existing_name(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.withColumn("fname", F.lit("blah"))

    dfs = employee.withColumn("fname", SF.lit("blah"))

    compare_frames(df, dfs, compare_schema=False)


def test_with_column_renamed(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.withColumnRenamed("fname", "first_name")

    dfs = employee.withColumnRenamed("fname", "first_name")

    compare_frames(df, dfs)


def test_with_column_renamed_double(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("fname").alias("first_name")).withColumnRenamed(
        "first_name", "first_name_again"
    )

    dfs = employee.select(SF.col("fname").alias("first_name")).withColumnRenamed(
        "first_name", "first_name_again"
    )

    compare_frames(df, dfs)


def test_with_columns(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.withColumns(
        {
            "test": F.col("age"),
            "test2": F.col("age"),
        }
    )

    dfs = employee.withColumns(
        {
            "test": SF.col("age"),
            "test2": SF.col("age"),
        }
    )

    compare_frames(df, dfs)


def test_with_columns_reference_another(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_bigquery: t.Callable,
    is_postgres: t.Callable,
):
    # Could consider two options:
    # 1. Use SQLGlot optimizer to properly change the references to be expanded to avoid the issue (a rule already does this)
    # 2. Write specific logic to both these dataframes to detect if there is a self reference and create a new scope
    if is_bigquery():
        pytest.skip(
            "BigQuery doesn't support having selects with columns that reference each other."
        )
    if is_postgres():
        pytest.skip(
            "Postgres doesn't support having selects with columns that reference each other."
        )
    employee = get_df("employee")
    df = pyspark_employee.withColumns(
        {
            "test": F.col("age"),
            "test2": F.col("test"),
        }
    )

    dfs = employee.withColumns(
        {
            "test": SF.col("age"),
            "test2": SF.col("test"),
        }
    )
    compare_frames(df, dfs)


def test_drop_column_single(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.select(F.col("fname"), F.col("lname"), F.col("age")).drop("age")

    dfs = employee.select(SF.col("fname"), SF.col("lname"), SF.col("age")).drop("age")

    compare_frames(df, dfs)


def test_drop_column_reference_join(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_spark_employee_cols = pyspark_employee.select(
        F.col("fname"), F.col("lname"), F.col("age"), F.col("store_id")
    )
    df_pyspark_store_cols = pyspark_store.select(F.col("store_id"), F.col("store_name"))
    df = df_spark_employee_cols.join(df_pyspark_store_cols, on="store_id", how="inner").drop(
        df_spark_employee_cols.age,
    )

    df_sqlglot_employee_cols = employee.select(
        SF.col("fname"), SF.col("lname"), SF.col("age"), SF.col("store_id")
    )
    df_sqlglot_store_cols = store.select(SF.col("store_id"), SF.col("store_name"))
    dfs = df_sqlglot_employee_cols.join(df_sqlglot_store_cols, on="store_id", how="inner").drop(
        df_sqlglot_employee_cols.age,
    )

    compare_frames(df, dfs, sort=True)


def test_limit(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.limit(1)

    dfs = employee.limit(1)

    compare_frames(df, dfs)


def test_hint_broadcast_alias(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    get_explain_plan: t.Callable,
    is_duckdb: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        pyspark_store.alias("store").hint("broadcast", "store"),
        on=pyspark_employee.store_id == pyspark_store.store_id,
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(
        store.alias("store").hint("broadcast", "store"),
        on=employee.store_id == store.store_id,
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
    )
    df, dfs = compare_frames(df_joined, dfs_joined)
    if isinstance(dfs, PySparkDataFrame):
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(df)
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(dfs)


def test_hint_broadcast_no_alias(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    get_explain_plan: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        pyspark_store.hint("broadcast"),
        on=pyspark_employee.store_id == pyspark_store.store_id,
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(
        store.hint("broadcast"),
        on=employee.store_id == store.store_id,
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
    )
    df, dfs = compare_frames(df_joined, dfs_joined)
    if isinstance(dfs, PySparkDataFrame):
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(df)
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(dfs)
        assert "UnresolvedHint BROADCAST, [" in get_explain_plan(dfs)


def test_broadcast_func(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    get_explain_plan: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_joined = pyspark_employee.join(
        F.broadcast(pyspark_store),
        on=pyspark_employee.store_id == pyspark_store.store_id,
        how="inner",
    ).select(
        pyspark_employee.employee_id,
        pyspark_employee["fname"],
        F.col("lname"),
        F.col("age"),
        pyspark_employee.store_id,
        pyspark_store.store_name,
        pyspark_store["num_sales"],
    )
    dfs_joined = employee.join(
        SF.broadcast(store),
        on=employee.store_id == store.store_id,
        how="inner",
    ).select(
        employee.employee_id,
        employee["fname"],
        SF.col("lname"),
        SF.col("age"),
        employee.store_id,
        store.store_name,
        store["num_sales"],
    )
    df, dfs = compare_frames(df_joined, dfs_joined)
    if isinstance(dfs, PySparkDataFrame):
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(df)
        assert "ResolvedHint (strategy=broadcast)" in get_explain_plan(dfs)
        assert "'UnresolvedHint BROADCAST, [" in get_explain_plan(dfs)


def test_repartition_by_num(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    """
    The results are different when doing the repartition on a table created using VALUES in SQL.
    So I just use the views instead for these tests
    """
    employee = get_df("employee")
    df = pyspark_employee.repartition(63)

    dfs = employee.session.read.table("employee").repartition(63)
    df, dfs = compare_frames(df, dfs)
    if isinstance(dfs, PySparkDataFrame):
        spark_num_partitions = df.rdd.getNumPartitions()
        sqlf_num_partitions = dfs.rdd.getNumPartitions()
        assert spark_num_partitions == 63
        assert spark_num_partitions == sqlf_num_partitions


def test_repartition_name_only(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    get_explain_plan: t.Callable,
):
    """
    We use the view here to help ensure the explain plans are similar enough to compare
    """
    employee = get_df("employee")
    df = pyspark_employee.repartition("age")

    dfs = employee.session.read.table("employee").repartition("age")
    df, dfs = compare_frames(df, dfs)
    if isinstance(dfs, PySparkDataFrame):
        assert "RepartitionByExpression [age" in get_explain_plan(df)
        assert "RepartitionByExpression [age" in get_explain_plan(dfs)


def test_repartition_num_and_multiple_names(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    get_explain_plan: t.Callable,
):
    """
    We use the view here to help ensure the explain plans are similar enough to compare
    """
    employee = get_df("employee")
    df = pyspark_employee.repartition(53, "age", "fname")

    dfs = employee.session.read.table("employee").repartition(53, "age", "fname")
    df, dfs = compare_frames(df, dfs)
    if isinstance(dfs, PySparkDataFrame):
        spark_num_partitions = df.rdd.getNumPartitions()
        sqlglot_num_partitions = dfs.rdd.getNumPartitions()
        assert spark_num_partitions == 53
        assert spark_num_partitions == sqlglot_num_partitions


def test_coalesce(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    df = pyspark_employee.coalesce(1)
    dfs = employee.coalesce(1)
    df, dfs = compare_frames(df, dfs)
    if isinstance(dfs, PySparkDataFrame):
        spark_num_partitions = df.rdd.getNumPartitions()
        sqlglot_num_partitions = dfs.rdd.getNumPartitions()
        assert spark_num_partitions == 1
        assert spark_num_partitions == sqlglot_num_partitions


def test_cache_select(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_employee = (
        pyspark_employee.groupBy("store_id")
        .agg(F.countDistinct("employee_id").alias("num_employees"))
        .cache()
    )
    df_joined = df_employee.join(pyspark_store, on="store_id").select(
        pyspark_store.store_id, df_employee.num_employees
    )
    dfs_employee = (
        employee.groupBy("store_id")
        .agg(SF.countDistinct("employee_id").alias("num_employees"))
        .cache()
    )
    dfs_joined = dfs_employee.join(store, on="store_id").select(
        store.store_id, dfs_employee.num_employees
    )
    compare_frames(df_joined, dfs_joined)


def test_persist_select(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    employee = get_df("employee")
    store = get_df("store")
    df_employee = (
        pyspark_employee.groupBy("store_id")
        .agg(F.countDistinct("employee_id").alias("num_employees"))
        .persist()
    )
    df_joined = df_employee.join(pyspark_store, on="store_id").select(
        pyspark_store.store_id, df_employee.num_employees
    )
    dfs_employee = (
        employee.groupBy("store_id")
        .agg(SF.countDistinct("employee_id").alias("num_employees"))
        .persist()
    )
    dfs_joined = dfs_employee.join(store, on="store_id").select(
        store.store_id, dfs_employee.num_employees
    )
    compare_frames(df_joined, dfs_joined)


def test_transform(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    def cast_all_to_int_pyspark(input_df):
        return input_df.select([F.col(col_name).cast("string") for col_name in input_df.columns])

    def cast_all_to_int_sqlframe(input_df):
        return input_df.select([SF.col(col_name).cast("string") for col_name in input_df.columns])

    def sort_columns_asc(input_df):
        return input_df.select(*sorted(input_df.columns))

    employee = get_df("employee")

    df = pyspark_employee.transform(cast_all_to_int_pyspark).transform(sort_columns_asc)
    dfs = employee.transform(cast_all_to_int_sqlframe).transform(sort_columns_asc)
    compare_frames(df, dfs)


# https://github.com/eakmanrq/sqlframe/issues/51
def test_join_full_outer_no_match(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    spark = pyspark_employee._session
    initial = spark.createDataFrame(
        [
            (1, "data"),
            (2, "data"),
        ],
        ["id", "data"],
    )
    for_join = spark.createDataFrame(
        [
            (1, "other_data"),
            (2, "other_data"),
            (3, "other_data"),
        ],
        ["id", "other_data"],
    )
    df = initial.join(for_join, on="id", how="full")

    session = get_df("employee").session
    dfs_initial = session.createDataFrame(
        [
            (1, "data"),
            (2, "data"),
        ],
        ["id", "data"],
    )
    dfs_for_join = session.createDataFrame(
        [
            (1, "other_data"),
            (2, "other_data"),
            (3, "other_data"),
        ],
        ["id", "other_data"],
    )
    dfs = dfs_initial.join(dfs_for_join, on="id", how="full")

    compare_frames(df, dfs)


# https://github.com/eakmanrq/sqlframe/issues/102
def test_join_with_duplicate_column_name(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    spark = pyspark_employee._session
    df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], ["age", "name"])
    df_height = spark.createDataFrame([(170, "Alice"), (180, "Bob")], ["height", "name"])

    df = df.join(df_height, how="left", on="name").withColumn("name", F.upper("name"))

    session = get_df("employee").session
    dfs = session.createDataFrame([(2, "Alice"), (5, "Bob")], ["age", "name"])
    dfs_height = session.createDataFrame([(170, "Alice"), (180, "Bob")], ["height", "name"])

    dfs = dfs.join(dfs_height, how="left", on="name").withColumn("name", SF.upper("name"))

    compare_frames(df, dfs, compare_schema=False)


# https://github.com/eakmanrq/sqlframe/issues/103
def test_chained_join_common_key(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
):
    spark = pyspark_employee._session
    df = spark.createDataFrame([(2, "Alice"), (5, "Bob")], ["age", "name"])
    height = spark.createDataFrame([(170, "Alice"), (180, "Bob")], ["height", "name"])
    location = spark.createDataFrame([("USA", "Alice"), ("Europe", "Bob")], ["location", "name"])

    df = df.join(height, how="left", on="name").join(location, how="left", on="name")

    session = get_df("employee").session
    dfs = session.createDataFrame([(2, "Alice"), (5, "Bob")], ["age", "name"])
    dfs_height = session.createDataFrame([(170, "Alice"), (180, "Bob")], ["height", "name"])
    dfs_location = session.createDataFrame(
        [("USA", "Alice"), ("Europe", "Bob")], ["location", "name"]
    )
    dfs = dfs.join(dfs_height, how="left", on="name").join(dfs_location, how="left", on="name")

    compare_frames(df, dfs, compare_schema=False)


# https://github.com/eakmanrq/sqlframe/issues/185
def test_chaining_joins_with_selects(
    pyspark_employee: PySparkDataFrame,
    pyspark_store: PySparkDataFrame,
    pyspark_district: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_spark: t.Callable,
):
    if is_spark():
        pytest.skip(
            "This test is not supported in Spark. This is related to how duplicate columns are handled in Spark"
        )
    df = (
        pyspark_employee.alias("employee")
        .join(
            pyspark_store.filter(F.col("store_name") != "test").alias("store"),
            on=F.col("employee.employee_id") == F.col("store.store_id"),
        )
        .join(
            pyspark_district.alias("district"),
            on=F.col("store.store_id") == F.col("district.district_id"),
        )
        .join(
            pyspark_district.alias("district2"),
            on=(F.col("store.store_id") + 1) == F.col("district2.district_id"),
            how="left",
        )
        .select("*")
    )

    employee = get_df("employee")
    store = get_df("store")
    district = get_df("district")

    dfs = (
        employee.alias("employee")
        .join(
            store.filter(SF.col("store_name") != "test").alias("store"),
            on=SF.col("employee.employee_id") == SF.col("store.store_id"),
        )
        .join(
            district.alias("district"),
            on=SF.col("store.store_id") == SF.col("district.district_id"),
        )
        .join(
            district.alias("district2"),
            on=(SF.col("store.store_id") + 1) == SF.col("district2.district_id"),
            how="left",
        )
        .select("*")
    )

    compare_frames(df, dfs, compare_schema=False)


# https://github.com/eakmanrq/sqlframe/issues/210
# https://github.com/eakmanrq/sqlframe/issues/212
def test_self_join(
    pyspark_employee: PySparkDataFrame,
    get_df: t.Callable[[str], _BaseDataFrame],
    compare_frames: t.Callable,
    is_spark: t.Callable,
):
    if is_spark():
        pytest.skip(
            "This test is not supported in Spark. This is related to how duplicate columns are handled in Spark"
        )
    df_filtered = pyspark_employee.where(F.col("age") > 40)
    df_joined = pyspark_employee.join(
        df_filtered,
        pyspark_employee["employee_id"].eqNullSafe(df_filtered["employee_id"]),
        how="inner",
    )

    employee = get_df("employee")

    dfs_filtered = employee.alias("dfs_filtered").where(SF.col("age") > 40)
    dfs_joined = employee.join(
        dfs_filtered,
        employee["employee_id"].eqNullSafe(dfs_filtered["employee_id"]),
        how="inner",
    )

    compare_frames(df_joined, dfs_joined, compare_schema=False)



import pytest

from sqlframe.base.catalog import CatalogMetadata, Column, Database, Function, Table
from sqlframe.duckdb.session import DuckDBSession


@pytest.fixture
def duckdb_session() -> DuckDBSession:
    import duckdb

    connector = duckdb.connect()
    connector.execute("set TimeZone = 'UTC'")
    connector.execute("SELECT * FROM duckdb_settings() WHERE name = 'TimeZone'")
    assert connector.fetchone()[1] == "UTC"  # type: ignore
    session = DuckDBSession(connector)
    session._execute('''ATTACH ':memory:' AS "default"''')
    session._execute('''ATTACH ':memory:' AS "catalog1"''')
    session._execute('''ATTACH ':memory:' AS "catalog2"''')
    session._execute('''USE "default"''')
    session._execute('CREATE SCHEMA "default"."db1"')
    session._execute('CREATE TABLE "default"."main"."table1" (id INTEGER, name VARCHAR(100))')
    session._execute("CREATE MACRO main.testing(a, b) AS a + b")
    return session


def test_current_catalog(duckdb_session: DuckDBSession):
    # Default is considered a word that must be quoted in DuckDB
    assert duckdb_session.catalog.currentCatalog() == "default"


def test_set_current_catalog(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.currentCatalog() == "default"
    duckdb_session.catalog.setCurrentCatalog("catalog1")
    assert duckdb_session.catalog.currentCatalog() == "catalog1"


def test_list_catalogs(duckdb_session: DuckDBSession):
    assert sorted(duckdb_session.catalog.listCatalogs(), key=lambda x: x.name) == [
        CatalogMetadata(name="catalog1", description=None),
        CatalogMetadata(name="catalog2", description=None),
        CatalogMetadata(name="default", description=None),
        CatalogMetadata(name="memory", description=None),
        CatalogMetadata(name="system", description=None),
        CatalogMetadata(name="temp", description=None),
    ]


def test_current_database(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.currentDatabase() == "main"


def test_set_current_database(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.currentDatabase() == "main"
    duckdb_session.catalog.setCurrentDatabase("db1")
    assert duckdb_session.catalog.currentDatabase() == "db1"


def test_list_databases(duckdb_session: DuckDBSession):
    assert sorted(duckdb_session.catalog.listDatabases(), key=lambda x: (x.catalog, x.name)) == [
        Database(name="information_schema", catalog="catalog1", description=None, locationUri=""),
        Database(name="main", catalog="catalog1", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="catalog1", description=None, locationUri=""),
        Database(name="information_schema", catalog="catalog2", description=None, locationUri=""),
        Database(name="main", catalog="catalog2", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="catalog2", description=None, locationUri=""),
        Database(name="db1", catalog="default", description=None, locationUri=""),
        Database(name="information_schema", catalog="default", description=None, locationUri=""),
        Database(name="main", catalog="default", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="default", description=None, locationUri=""),
        Database(name="information_schema", catalog="memory", description=None, locationUri=""),
        Database(name="main", catalog="memory", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="memory", description=None, locationUri=""),
        Database(name="information_schema", catalog="system", description=None, locationUri=""),
        Database(name="main", catalog="system", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="system", description=None, locationUri=""),
        Database(name="information_schema", catalog="temp", description=None, locationUri=""),
        Database(name="main", catalog="temp", description=None, locationUri=""),
        Database(name="pg_catalog", catalog="temp", description=None, locationUri=""),
    ]


def test_list_databases_pattern(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listDatabases("db*"), key=lambda x: (x.catalog, x.name)
    ) == [
        Database(name="db1", catalog="default", description=None, locationUri=""),
    ]


def test_get_database_no_match(duckdb_session: DuckDBSession):
    with pytest.raises(ValueError):
        assert duckdb_session.catalog.getDatabase("nonexistent")


def test_get_database_name_only(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.getDatabase("db1") == Database(
        name="db1", catalog="default", description=None, locationUri=""
    )


def test_get_database_name_and_catalog(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.getDatabase("catalog1.information_schema") == Database(
        name="information_schema", catalog="catalog1", description=None, locationUri=""
    )


def test_database_exists_does_exist(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.databaseExists("db1") is True


def test_database_exists_does_not_exist(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.databaseExists("nonexistent") is False


def test_list_tables_no_args(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listTables(), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="default",
            namespace=["main"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_no_catalog(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listTables("main"), key=lambda x: (x.catalog, x.database, x.name)
    ) == [
        Table(
            name="table1",
            catalog="default",
            namespace=["main"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_db_and_catalog(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listTables("default.main"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="default",
            namespace=["main"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_list_tables_pattern(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listTables(pattern="tab*"),
        key=lambda x: (x.catalog, x.database, x.name),
    ) == [
        Table(
            name="table1",
            catalog="default",
            namespace=["main"],
            description=None,
            tableType="MANAGED",
            isTemporary=False,
        )
    ]


def test_get_table(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.getTable("default.main.table1") == Table(
        name="table1",
        catalog="default",
        namespace=["main"],
        description=None,
        tableType="MANAGED",
        isTemporary=False,
    )


def test_get_table_not_exists(duckdb_session: DuckDBSession):
    with pytest.raises(ValueError):
        assert duckdb_session.catalog.getTable("default.main.nonexistent")


def test_list_functions(duckdb_session: DuckDBSession):
    assert (
        Function(
            name="testing",
            catalog="default",
            namespace=["main"],
            description=None,
            className="",
            isTemporary=False,
        )
        in duckdb_session.catalog.listFunctions()
    )


def test_list_functions_pattern(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.listFunctions(pattern="test*") == [
        Function(
            name="testing",
            catalog="default",
            namespace=["main"],
            description=None,
            className="",
            isTemporary=False,
        )
    ]


def test_function_exists_does_exist(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.functionExists("testing") is True


def test_function_exists_does_not_exist(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.functionExists("nonexistent") is False


def test_get_function_exists(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.getFunction("default.main.testing") == Function(
        name="testing",
        catalog="default",
        namespace=["main"],
        description=None,
        className="",
        isTemporary=False,
    )


def test_get_function_not_exists(duckdb_session: DuckDBSession):
    with pytest.raises(ValueError):
        assert duckdb_session.catalog.getFunction("default.main.nonexistent")


def test_list_columns(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listColumns("default.main.table1"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_list_columns_use_db_name(duckdb_session: DuckDBSession):
    assert sorted(
        duckdb_session.catalog.listColumns("table1", dbName="default.main"), key=lambda x: x.name
    ) == [
        Column(
            name="id",
            description=None,
            dataType="INT",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
        Column(
            name="name",
            description=None,
            dataType="STRING",
            nullable=True,
            isPartition=False,
            isBucket=False,
        ),
    ]


def test_table_exists_table_name_only(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.tableExists("default.main.table1") is True


def test_table_exists_table_name_and_db_name(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.tableExists("table1", dbName="default.main") is True


def test_table_not_exists(duckdb_session: DuckDBSession):
    assert duckdb_session.catalog.tableExists("nonexistent") is False


def test_create_external_table(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.createExternalTable("table1", "default.main", "path/to/table")


def test_create_table(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.createTable("table1", "default.main")


def test_drop_temp_view(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.dropTempView("view1")


def test_drop_global_temp_view(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.dropGlobalTempView("view1")


def test_register_function(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.registerFunction("function1", lambda x: x)


def test_is_cached(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.isCached("table1")


def test_cache_table(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.cacheTable("table1")


def test_uncache_table(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.uncacheTable("table1")


def test_clear_cache(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.clearCache()


def test_refresh_table(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.refreshTable("table1")


def test_recover_partitions(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.recoverPartitions("table1")


def test_refresh_by_path(duckdb_session: DuckDBSession):
    with pytest.raises(NotImplementedError):
        duckdb_session.catalog.refreshByPath("path/to/table")
